# Copyright (c) OpenMMLab. All rights reserved.
import copy
import math
import os
import random
from abc import abstractmethod
from typing import Dict, List, Optional, Sequence, Tuple, Union

import torch
from mmengine import fileio
from mmengine.evaluator import Evaluator
from mmengine.runner import IterBasedTrainLoop
from mmengine.utils import is_list_of
from torch.utils.data import DataLoader

from mmrazor.models.task_modules.estimators import get_model_complexity_info
from mmrazor.registry import LOOPS
from mmrazor.structures import Candidates, export_fix_subnet, load_fix_subnet
from mmrazor.utils import SupportRandomSubnet


class BaseSamplerTrainLoop(IterBasedTrainLoop):
    """IterBasedTrainLoop for base sampler.

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or dict): A dataloader object or a dict to
            build a dataloader for training the model.
        max_iters (int): Total training iters.
        val_begin (int): The iteration that begins validating.
            Defaults to 1.
        val_interval (int): Validation interval. Defaults to 1000.
    """

    def __init__(self,
                 runner,
                 dataloader: Union[Dict, DataLoader],
                 max_iters: int,
                 val_begin: int = 1,
                 val_interval: int = 1000):
        super().__init__(runner, dataloader, max_iters, val_begin,
                         val_interval)
        if self.runner.distributed:
            self.model = runner.model.module
        else:
            self.model = runner.model

    @abstractmethod
    def sample_subnet(self) -> SupportRandomSubnet:
        """Sample a subnet to train the supernet."""

    def run_iter(self, data_batch: Sequence[dict]) -> None:
        """Iterate one mini-batch.

        Args:
            data_batch (Sequence[dict]): Batch of data from dataloader.
        """
        self.runner.call_hook(
            'before_train_iter', batch_idx=self._iter, data_batch=data_batch)
        # Enable gradient accumulation mode and avoid unnecessary gradient
        # synchronization during gradient accumulation process.
        # outputs should be a dict of loss.
        subnet = self.sample_subnet()
        self.model.set_subnet(subnet)
        outputs = self.runner.model.train_step(
            data_batch, optim_wrapper=self.runner.optim_wrapper)
        self.runner.message_hub.update_info('train_logs', outputs)

        self.runner.call_hook(
            'after_train_iter',
            batch_idx=self._iter,
            data_batch=data_batch,
            outputs=outputs)
        self._iter += 1


@LOOPS.register_module()
class GreedySamplerTrainLoop(BaseSamplerTrainLoop):
    """IterBasedTrainLoop for greedy sampler.

    In GreedySamplerTrainLoop, `Greedy` means that only use some top
    sampled candidates to train the supernet. So GreedySamplerTrainLoop mainly
    picks the top candidates based on their val socres, then use them to train
    the supernet one by one.

    Steps:
        1. Sample from the supernet and the candidates.
        2. Validate these sampled candidates to get each candidate's score.
        3. Get top-k candidates based on their scores, then use them to train
        the supernet one by one.

    Args:
        runner (Runner): A reference of runner.
        dataloader (Dataloader or dict): A dataloader object or a dict to
            build a dataloader for training the model.
        dataloader_val (Dataloader or dict): A dataloader object or a dict to
            build a dataloader for evaluating the candidates.
        evaluator (Evaluator or dict or list): Used for computing metrics.
        max_iters (int): Total training iters.
        val_begin (int): The iteration that begins validating.
            Defaults to 1.
        val_interval (int): Validation interval. Defaults to 1000.
        score_key (str): Specify one metric in evaluation results to score
            candidates. Defaults to 'accuracy_top-1'.
        flops_range (dict): Constraints to be used for screening candidates.
        spec_modules (list): Used for specify modules need to counter.
            Defaults to list().
        num_candidates (int): The number of the candidates consist of samples
            from supernet and itself. Defaults to 1000.
        num_samples (int): The number of sample in each sampling subnet.
            Defaults to 10.
        top_k (int): Choose top_k subnet from the candidates used to train
            the supernet. Defaults to 5.
        prob_schedule (str): The schedule to generate the probablity of
            sampling from the candidates. The probablity will increase from
            [init_prob, max_prob] during [schedule_start_iter,
            schedule_end_iter]. Both of 'linear' schedule and 'consine'
            schedule are supported. Defaults to 'linear'.
        schedule_start_iter (int): The start iter of the prob_schedule.
            Defaults to 10000. 10000 is corresponding to batch_size: 1024.
            You should adptive it based on your batch_size.
        schedule_end_iter (int): The end iter in of the prob_schedule.
            Defaults to 144360. 144360 = 120(epoch) * 1203 (iters/epoch),
            batch_size is 1024. You should adptive it based on the batch_size
            and the total training epochs.
        init_prob (float): The init probablity of the prob_schedule.
            Defaults to 0.0.
        max_prob (float): The max probablity of the prob_schedule.
            Defaults to 0.8.
    """

    def __init__(self,
                 runner,
                 dataloader: Union[Dict, DataLoader],
                 dataloader_val: Union[Dict, DataLoader],
                 evaluator: Union[Evaluator, Dict, List],
                 max_iters: int,
                 val_begin: int = 1,
                 val_interval: int = 1000,
                 score_key: str = 'accuracy/top1',
                 flops_range: Optional[Tuple[float, float]] = (0., 330 * 1e6),
                 spec_modules: List = [],
                 num_candidates: int = 1000,
                 num_samples: int = 10,
                 top_k: int = 5,
                 prob_schedule: str = 'linear',
                 schedule_start_iter: int = 10000,
                 schedule_end_iter: int = 144360,
                 init_prob: float = 0.,
                 max_prob: float = 0.8) -> None:
        super().__init__(runner, dataloader, max_iters, val_begin,
                         val_interval)
        if isinstance(dataloader_val, dict):
            self.dataloader_val = runner.build_dataloader(
                dataloader_val, seed=runner.seed)
        else:
            self.dataloader_val = dataloader_val

        if isinstance(evaluator, dict) or is_list_of(evaluator, dict):
            self.evaluator = runner.build_evaluator(evaluator)
        else:
            self.evaluator = evaluator

        self.score_key = score_key
        self.flops_range = flops_range
        self.spec_modules = spec_modules
        self.num_candidates = num_candidates
        self.num_samples = num_samples
        self.top_k = top_k
        assert prob_schedule in ['linear', 'consine']
        self.prob_schedule = prob_schedule
        self.schedule_start_iter = schedule_start_iter
        self.schedule_end_iter = schedule_end_iter
        self.init_prob = init_prob
        self.max_prob = max_prob
        self.cur_prob: float = 0.

        self.candidates = Candidates()
        self.top_k_candidates = Candidates()

    def run(self) -> None:
        """Launch training."""
        self.runner.call_hook('before_train')
        # In iteration-based training loop, we treat the whole training process
        # as a big epoch and execute the corresponding hook.
        self.runner.call_hook('before_train_epoch')
        while self._iter < self._max_iters:
            self.runner.model.train()

            data_batch = next(self.dataloader_iterator)
            self.run_iter(data_batch)

            if (self.runner.val_loop is not None
                    and self._iter >= self.runner.val_begin
                    and self._iter % self.runner.val_interval == 0):
                self.runner.val_loop.run()
        self._save_candidates()

        self.runner.call_hook('after_train_epoch')
        self.runner.call_hook('after_train')

    def sample_subnet(self) -> SupportRandomSubnet:
        """Sample a subnet from top_k candidates one by one, then to train the
        surpernet with the subnet.

        Steps:
            1. Update and get the `top_k_candidates`.
                1.1. Update the prob of sampling from the `candidates` based on
                the `prob_schedule` and the current iter.
                1.2. Sample `num_samples` candidates from the supernet and the
                `candidates` based on the updated prob(step 1.1).
                1.3. Val all candidates to get their scores, including the
                sampled candidates(step 1.2).
                1.4. Update the `top_k_candidates` based on
                their scores(step 1.3).
            2. Pop from the `top_k_candidates` one by one to train
            the supernet.
        """
        if len(self.top_k_candidates) == 0:
            self.update_cur_prob(cur_iter=self._iter)

            sampled_candidates, num_sample_from_supernet = \
                self.get_candidates_with_sample(num_samples=self.num_samples)

            self.candidates.extend(sampled_candidates)

            self.update_candidates_scores()

            self.candidates.sort(key=lambda x: x[1], reverse=True)
            self.candidates = Candidates(self.candidates[:self.num_candidates])
            self.top_k_candidates = Candidates(self.candidates[:self.top_k])

            top1_score = self.top_k_candidates.scores[0]
            if (self._iter % self.val_interval) < self.top_k:
                self.runner.logger.info(
                    f'GreedySampler: [{self._iter:>6d}] '
                    f'prob {self.cur_prob:.3f} '
                    f'num_sample_from_supernet '
                    f'{num_sample_from_supernet}/{self.num_samples} '
                    f'top1_score {top1_score:.3f} '
                    f'cur_num_candidates: {len(self.candidates)}')
        return self.top_k_candidates.pop(0)[0]

    def update_cur_prob(self, cur_iter: int) -> None:
        """update current probablity of sampling from the candidates, which is
        generated based on the probablity strategy and current iter."""
        if cur_iter > self.schedule_end_iter:
            self.cur_prob = self.max_prob
        elif cur_iter < self.schedule_start_iter:
            self.cur_prob = self.init_prob
        else:
            schedule_all_steps = self.schedule_end_iter - \
                self.schedule_start_iter
            schedule_cur_steps = cur_iter - self.schedule_start_iter
            if self.prob_schedule == 'linear':
                tmp = self.max_prob - self.init_prob
                self.cur_prob = tmp / schedule_all_steps * schedule_cur_steps
            elif self.prob_schedule == 'consine':
                tmp_1 = (1 - self.init_prob) * 0.5
                tmp_2 = math.pi * schedule_cur_steps
                tmp_3 = schedule_all_steps
                self.cur_prob = tmp_1 * (1 + math.cos(tmp_2 / tmp_3)) \
                    + self.init_prob
            else:
                raise ValueError('`prob_schedule` is eroor, it should be \
                    one of `linear` and `consine`.')

    def get_candidates_with_sample(self,
                                   num_samples: int) -> Tuple[Candidates, int]:
        """Get candidates with sampling from supernet and the candidates based
        on the current probablity."""
        num_sample_from_supernet = 0
        sampled_candidates = Candidates()
        for _ in range(num_samples):
            if random.random() >= self.cur_prob or len(self.candidates) == 0:
                subnet = self._sample_from_supernet()
                if self._check_constraints(subnet):
                    sampled_candidates.append(subnet)
                num_sample_from_supernet += 1
            else:
                sampled_candidates.append(self._sample_from_candidates())
        return sampled_candidates, num_sample_from_supernet

    def update_candidates_scores(self) -> None:
        """Update candidates' scores, which are validated with the
        `dataloader_val`."""
        for i, candidate in enumerate(self.candidates.subnets):
            self.model.set_subnet(candidate)
            metrics = self._val_candidate()
            score = metrics[self.score_key] if len(metrics) != 0 else 0.
            self.candidates.set_score(i, score)

    @torch.no_grad()
    def _val_candidate(self) -> Dict:
        """Run validation."""
        self.runner.model.eval()
        for data_batch in self.dataloader_val:
            outputs = self.runner.model.val_step(data_batch)
            self.evaluator.process(data_samples=outputs, data_batch=data_batch)
        metrics = self.evaluator.evaluate(len(self.dataloader_val.dataset))
        return metrics

    def _sample_from_supernet(self) -> SupportRandomSubnet:
        """Sample from the supernet."""
        subnet = self.model.sample_subnet()
        return subnet

    def _sample_from_candidates(self) -> SupportRandomSubnet:
        """Sample from the candidates."""
        assert len(self.candidates) > 0
        subnet = random.choice(self.candidates)
        return subnet

    def _check_constraints(self, random_subnet: SupportRandomSubnet) -> bool:
        """Check whether is beyond constraints.

        Returns:
            bool: The result of checking.
        """
        if self.flops_range is None:
            return True

        self.model.set_subnet(random_subnet)
        fix_mutable = export_fix_subnet(self.model)
        copied_model = copy.deepcopy(self.model)
        load_fix_subnet(copied_model, fix_mutable)
        flops, _ = get_model_complexity_info(
            copied_model, spec_modules=self.spec_modules)

        if self.flops_range[0] <= flops <= self.flops_range[1]:
            return True
        else:
            return False

    def _save_candidates(self) -> None:
        """Save the candidates to init the next searching."""
        save_path = os.path.join(self.runner.work_dir, 'candidates.pkl')
        fileio.dump(self.candidates, save_path)
        self.runner.logger.info(f'candidates.pkl saved in '
                                f'{self.runner.work_dir}')



_base_ = [
    'mmrazor::_base_/settings/cifar10_darts_subnet.py',
    'mmrazor::_base_/nas_backbones/darts_supernet.py',
    'mmcls::_base_/default_runtime.py',
]

subnet_backbone = _base_.nas_backbone
subnet_backbone.base_channels = 36
subnet_backbone.num_layers = 20
subnet_backbone.auxliary = True
subnet_backbone.aux_channels = 128
subnet_backbone.aux_out_channels = 768
subnet_backbone.out_indices = (19, )
subnet_backbone.norm_cfg = norm_cfg = dict(type='BN', affine=True)

# model
supernet = dict(
    type='ImageClassifier',
    data_preprocessor=_base_.preprocess_cfg,
    backbone=subnet_backbone,
    neck=dict(type='GlobalAveragePooling'),
    head=dict(
        type='mmrazor.DartsSubnetClsHead',
        num_classes=10,
        in_channels=576,
        aux_in_channels=768,
        loss=dict(type='CrossEntropyLoss', loss_weight=1.0),
        aux_loss=dict(type='CrossEntropyLoss', loss_weight=0.4),
        topk=(1, 5),
        cal_acc=True))

fix_subnet = 'configs/nas/mmcls/darts/DARTS_SUBNET_CIFAR_PAPER_ALIAS.yaml'

model = dict(
    type='mmrazor.SPOS',
    architecture=supernet,
    mutator=None,
    fix_subnet=fix_subnet,
)

find_unused_parameter = False



# Copyright (c) OpenMMLab. All rights reserved.
from abc import abstractmethod
from typing import List

import torch

from ..base_mutable import CHOICE_TYPE, CHOSEN_TYPE, BaseMutable
from ..derived_mutable import DerivedMethodMixin


class MutableChannel(BaseMutable[CHOICE_TYPE, CHOSEN_TYPE],
                     DerivedMethodMixin):
    """A type of ``MUTABLES`` for single path supernet such as AutoSlim. In
    single path supernet, each module only has one choice invoked at the same
    time. A path is obtained by sampling all the available choices. It is the
    base class for one shot channel mutables.

    Args:
        num_channels (int): The raw number of channels.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(self, num_channels: int, **kwargs):
        super().__init__(**kwargs)

        self.num_channels = num_channels
        self._same_mutables: List[MutableChannel] = list()

        # If the input of a module is a concatenation of several modules'
        # outputs, we add the mutable out of these modules to the
        # `concat_parent_mutables` of this module.
        self.concat_parent_mutables: List[MutableChannel] = list()
        self.name = 'unbind'

    @property
    def same_mutables(self):
        """Mutables in `same_mutables` and the current mutable should change
        Synchronously."""
        return self._same_mutables

    def register_same_mutable(self, mutable):
        """Register the input mutable in `same_mutables`."""
        if isinstance(mutable, list):
            # Add a concatenation of mutables to `concat_parent_mutables`.
            self.concat_parent_mutables = mutable
            return

        if self == mutable:
            return
        if mutable in self._same_mutables:
            return

        self._same_mutables.append(mutable)
        for s_mutable in self._same_mutables:
            s_mutable.register_same_mutable(mutable)
            mutable.register_same_mutable(s_mutable)

    @abstractmethod
    def convert_choice_to_mask(self, choice: CHOICE_TYPE) -> torch.Tensor:
        """Get the mask according to the input choice."""
        pass

    @property
    def current_mask(self):
        """The current mask.

        We slice the registered parameters and buffers of a ``nn.Module``
        according to the mask of the corresponding channel mutable.
        """
        if len(self.concat_parent_mutables) > 0:
            # If the input of a module is a concatenation of several modules'
            # outputs, the in_mask of this module is the concatenation of
            # these modules' out_mask.
            return torch.cat([
                mutable.current_mask for mutable in self.concat_parent_mutables
            ])
        else:
            return self.convert_choice_to_mask(self.current_choice)

    def bind_mutable_name(self, name: str) -> None:
        """Bind a MutableChannel to its name.

        Args:
            name (str): Name of this `MutableChannel`.
        """
        self.name = name

    def fix_chosen(self, chosen: CHOSEN_TYPE) -> None:
        """Fix mutable with subnet config. This operation would convert
        `unfixed` mode to `fixed` mode. The :attr:`is_fixed` will be set to
        True and only the selected operations can be retained.

        Args:
            chosen (str): The chosen key in ``MUTABLE``. Defaults to None.
        """
        if self.is_fixed:
            raise AttributeError(
                'The mode of current MUTABLE is `fixed`. '
                'Please do not call `fix_chosen` function again.')

        self.is_fixed = True

    def __repr__(self):
        concat_mutable_name = [
            mutable.name for mutable in self.concat_parent_mutables
        ]
        repr_str = self.__class__.__name__
        repr_str += f'(name={self.name}, '
        repr_str += f'num_channels={self.num_channels}, '
        repr_str += f'concat_mutable_name={concat_mutable_name})'
        return repr_str



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from abc import abstractmethod
from typing import Dict, List, Optional

from torch.nn import Module

from mmrazor.registry import MODELS, TASK_UTILS
from ...mutables import MutableChannel
from ...task_modules import PathConcatNode, PathDepthWiseConvNode, PathList
from ..base_mutator import BaseMutator
from ..utils import DEFAULT_MODULE_CONVERTERS


@MODELS.register_module()
class ChannelMutator(BaseMutator):
    """Base class for channel-based mutators.

    Args:
        mutable_cfg (dict): The config for the channel mutable.
        tracer_cfg (dict | Optional): The config for the model tracer.
            We Trace the topology of a given model with the tracer.
        skip_prefixes (List[str] | Optional): The module whose name start with
            a string in skip_prefixes will not be pruned.
        init_cfg (dict, optional): The config to control the initialization.

    Attributes:
        search_groups (Dict[int, List]): Search group of supernet. Note that
            the search group of a mutable based channel mutator is composed of
            corresponding mutables. Mutables in the same search group should
            be pruned together.
        name2module (Dict[str, :obj:`torch.nn.Module`]): The mapping from
            a module name to the module.

    Notes:
        # To avoid ambiguity, we only allow the following two cases:
        # 1. None of the parent nodes of a node is a `ConcatNode`
        # 2. A node has only one parent node which is a `ConcatNode`
    """

    def __init__(
        self,
        mutable_cfg: Dict,
        tracer_cfg: Optional[Dict] = None,
        skip_prefixes: Optional[List[str]] = None,
        init_cfg: Optional[Dict] = None,
    ) -> None:
        super().__init__(init_cfg)

        self.mutable_cfg = mutable_cfg
        if tracer_cfg:
            self.tracer = TASK_UTILS.build(tracer_cfg)
        else:
            self.tracer = None
        self.skip_prefixes = skip_prefixes
        self._search_groups: Optional[Dict[int, List[Module]]] = None

    def add_link(self, path_list: PathList) -> None:
        """Establish the relationship between the current nodes and their
        parents."""
        for path in path_list:
            pre_node = None
            for node in path:
                if isinstance(node, PathDepthWiseConvNode):
                    module = self.name2module[node.name]
                    # The in_channels and out_channels of a depth-wise conv
                    # should be the same
                    module.mutable_out.register_same_mutable(module.mutable_in)
                    module.mutable_in.register_same_mutable(module.mutable_out)

                if isinstance(node, PathConcatNode):
                    if pre_node is not None:
                        module_names = node.get_module_names()
                        concat_modules = [
                            self.name2module[name] for name in module_names
                        ]
                        concat_mutables = [
                            module.mutable_out for module in concat_modules
                        ]
                        pre_module = self.name2module[pre_node.name]
                        pre_module.mutable_in.register_same_mutable(
                            concat_mutables)

                    for sub_path_list in node:
                        self.add_link(sub_path_list)

                    # ConcatNode is the last node in a path
                    break

                if pre_node is None:
                    pre_node = node
                    continue

                pre_module = self.name2module[pre_node.name]
                cur_module = self.name2module[node.name]
                pre_module.mutable_in.register_same_mutable(
                    cur_module.mutable_out)
                cur_module.mutable_out.register_same_mutable(
                    pre_module.mutable_in)

                pre_node = node

    def prepare_from_supernet(self, supernet: Module) -> None:
        """Do some necessary preparations with supernet.

        We support the following two cases:

        Case 1: The input is the original nn.Module. We first replace the
        conv/linear/norm modules in the input supernet with dynamic ops.
        And trace the topology of the supernet. Finally, `search_groups` can be
        built based on the topology.

        Case 2: The input supernet is made up of dynamic ops. In this case,
        relationship between nodes and their parents must have been
        established and topology of the supernet is available for us. Then
        `search_groups` can be built based on the topology.

        Args:
            supernet (:obj:`torch.nn.Module`): The supernet to be searched
                in your algorithm.
        """

        if self.tracer is not None:
            self.convert_dynamic_module(supernet, self.module_converters)
            # The mapping from a module name to the module
            self._name2module = dict(supernet.named_modules())

            assert self.tracer is not None
            module_path_list: PathList = self.tracer.trace(supernet)

            self.add_link(module_path_list)
        else:
            self._name2module = dict(supernet.named_modules())

        self.bind_mutable_name(supernet)
        self._search_groups = self.build_search_groups(supernet)

    @staticmethod
    def find_same_mutables(supernet) -> Dict:
        """The mutables in the same group should be pruned together."""
        visited = []
        groups = {}
        group_idx = 0
        for name, module in supernet.named_modules():
            if isinstance(module, MutableChannel):
                same_mutables = module.same_mutables
                if module not in visited and len(same_mutables) > 0:
                    groups[group_idx] = [module] + same_mutables
                    visited.extend(groups[group_idx])
                    group_idx += 1
        return groups

    def bind_mutable_name(self, supernet: Module):
        """Bind a MutableChannel to its name.

        Args:
            supernet (:obj:`torch.nn.Module`): The supernet to be searched
                in your algorithm.
        """

        def traverse(module, prefix):
            for name, child in module.named_children():
                module_name = f'{prefix}.{name}' if prefix else name

                if isinstance(child, MutableChannel):
                    child.bind_mutable_name(prefix)
                else:
                    traverse(child, module_name)

        traverse(supernet, '')

    def convert_dynamic_module(self, supernet: Module, converters: Dict):
        """Replace the conv/linear/norm modules in the input supernet with
        dynamic ops.

        Args:
            supernet (:obj:`torch.nn.Module`): The architecture to be converted
                in your algorithm.
            dynamic_layer (Dict): The mapping from the module type to the
                corresponding dynamic layer.
        """

        def traverse(module, prefix):
            for name, child in module.named_children():
                module_name = prefix + name

                if type(child) in converters:
                    mutable_cfg_ = copy.deepcopy(self.mutable_cfg)
                    converter = converters[type(child)]
                    layer = converter(child, mutable_cfg_, mutable_cfg_)
                    setattr(module, name, layer)
                else:
                    traverse(child, module_name + '.')

        traverse(supernet, '')

    @abstractmethod
    def build_search_groups(self, supernet: Module):
        """Build `search_groups`.

        The mutables in the same group should be pruned together.
        """

    @property
    def search_groups(self) -> Dict[int, List]:
        """Search group of supernet.

        Note:
            For mutable based mutator, the search group is composed of
            corresponding mutables.

        Raises:
            RuntimeError: Called before search group has been built.

        Returns:
            Dict[int, List[MUTABLE_TYPE]]: Search group.
        """
        if self._search_groups is None:
            raise RuntimeError(
                'Call `search_groups` before access `build_search_groups`!')
        return self._search_groups

    @property
    def name2module(self):
        """The mapping from a module name to the module.

        Returns:
            dict: The name to module mapping.
        """
        if hasattr(self, '_name2module'):
            return self._name2module
        else:
            raise RuntimeError('Called before access `prepare_from_supernet`!')

    @property
    def module_converters(self) -> Dict:
        """The mapping from a type to the corresponding dynamic layer. It is
        called in `prepare_from_supernet`.

        Returns:
            dict: The mapping dict.
        """
        return DEFAULT_MODULE_CONVERTERS

    def is_skip_pruning(self, module_name: str,
                        skip_prefixes: Optional[List[str]]) -> bool:
        """Judge if the module with the input `module_name` should not be
        pruned.

        Args:
            module_name (str): Module name.
            skip_prefixes (list or None): The module whose name start with
                a string in skip_prefixes will not be prune.
        """
        skip_pruning = False
        if skip_prefixes:
            for prefix in skip_prefixes:
                if module_name.startswith(prefix):
                    skip_pruning = True
                    break
        return skip_pruning



_base_ = './detnas_shufflenet_supernet_8xb128_in1k.py'

# FIXME: you may replace this with the mutable_cfg searched by yourself
fix_subnet = 'https://download.openmmlab.com/mmrazor/v1/detnas/detnas_subnet_frcnn_shufflenetv2_fpn_1x_coco_bbox_backbone_flops-0.34M_mAP-37.5_20220715-61d2e900_subnet_cfg_v1.yaml'  # noqa: E501

model = dict(fix_subnet=fix_subnet)

find_unused_parameters = False



mutable_cfg = dict(
    type='mmrazor.DiffMutableOP',
    candidates=dict(
        zero=dict(type='mmrazor.DartsZero'),
        skip_connect=dict(type='mmrazor.DartsSkipConnect', use_drop_path=True),
        max_pool_3x3=dict(
            type='mmrazor.DartsPoolBN', pool_type='max', use_drop_path=True),
        avg_pool_3x3=dict(
            type='mmrazor.DartsPoolBN', pool_type='avg', use_drop_path=True),
        sep_conv_3x3=dict(
            type='mmrazor.DartsSepConv', kernel_size=3, use_drop_path=True),
        sep_conv_5x5=dict(
            type='mmrazor.DartsSepConv', kernel_size=5, use_drop_path=True),
        dil_conv_3x3=dict(
            type='mmrazor.DartsDilConv', kernel_size=3, use_drop_path=True),
        dil_conv_5x5=dict(
            type='mmrazor.DartsDilConv', kernel_size=5, use_drop_path=True)))

route_cfg = dict(type='mmrazor.DiffChoiceRoute', with_arch_param=True)

nas_backbone = dict(
    type='mmrazor.DartsBackbone',
    in_channels=3,
    base_channels=16,
    num_layers=8,
    num_nodes=4,
    stem_multiplier=3,
    out_indices=(7, ),
    mutable_cfg=mutable_cfg,
    route_cfg=route_cfg,
    norm_cfg=dict(type='BN', affine=False))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Callable, Dict, Optional

from torch import nn
from torch.nn.modules import GroupNorm
from torch.nn.modules.batchnorm import _BatchNorm
from torch.nn.modules.instancenorm import _InstanceNorm

from ...architectures import (DynamicBatchNorm, DynamicConv2d,
                              DynamicGroupNorm, DynamicInstanceNorm,
                              DynamicLinear)


def dynamic_conv2d_converter(module: nn.Conv2d, in_channels_cfg: Dict,
                             out_channels_cfg: Dict) -> DynamicConv2d:
    """Convert a nn.Conv2d module to a DynamicConv2d.

    Args:
        module (:obj:`torch.nn.Conv2d`): The original Conv2d module.
        in_channels_cfg (Dict): Config related to `in_channels`.
        out_channels_cfg (Dict): Config related to `out_channels`.
    """
    dynamic_conv = DynamicConv2d(
        in_channels_cfg=in_channels_cfg,
        out_channels_cfg=out_channels_cfg,
        in_channels=module.in_channels,
        out_channels=module.out_channels,
        kernel_size=module.kernel_size,
        stride=module.stride,
        padding=module.padding,
        dilation=module.dilation,
        groups=module.groups,
        bias=True if module.bias is not None else False,
        padding_mode=module.padding_mode)
    return dynamic_conv


def dynamic_linear_converter(module: nn.Linear, in_channels_cfg: Dict,
                             out_channels_cfg: Dict) -> DynamicLinear:
    """Convert a nn.Linear module to a DynamicLinear.

    Args:
        module (:obj:`torch.nn.Linear`): The original Linear module.
        in_features_cfg (Dict): Config related to `in_features`.
        out_features_cfg (Dict): Config related to `out_features`.
    """
    dynamic_linear = DynamicLinear(
        in_features_cfg=in_channels_cfg,
        out_features_cfg=out_channels_cfg,
        in_features=module.in_features,
        out_features=module.out_features,
        bias=True if module.bias is not None else False)
    return dynamic_linear


def dynamic_bn_converter(
        module: _BatchNorm,
        in_channels_cfg: Dict,
        out_channels_cfg: Optional[Dict] = None) -> DynamicBatchNorm:
    """Convert a _BatchNorm module to a DynamicBatchNorm.

    Args:
        module (:obj:`torch.nn._BatchNorm`): The original BatchNorm module.
        num_features_cfg (Dict): Config related to `num_features`.
    """
    dynamic_bn = DynamicBatchNorm(
        num_features_cfg=in_channels_cfg,
        num_features=module.num_features,
        eps=module.eps,
        momentum=module.momentum,
        affine=module.affine,
        track_running_stats=module.track_running_stats)
    return dynamic_bn


def dynamic_in_converter(
        module: _InstanceNorm,
        in_channels_cfg: Dict,
        out_channels_cfg: Optional[Dict] = None) -> DynamicInstanceNorm:
    """Convert a _InstanceNorm module to a DynamicInstanceNorm.

    Args:
        module (:obj:`torch.nn._InstanceNorm`): The original InstanceNorm
            module.
        num_features_cfg (Dict): Config related to `num_features`.
    """
    dynamic_in = DynamicInstanceNorm(
        num_features_cfg=in_channels_cfg,
        num_features=module.num_features,
        eps=module.eps,
        momentum=module.momentum,
        affine=module.affine,
        track_running_stats=module.track_running_stats)
    return dynamic_in


def dynamic_gn_converter(
        module: GroupNorm,
        in_channels_cfg: Dict,
        out_channels_cfg: Optional[Dict] = None) -> DynamicGroupNorm:
    """Convert a GroupNorm module to a DynamicGroupNorm.

    Args:
        module (:obj:`torch.nn.GroupNorm`): The original GroupNorm module.
        num_channels_cfg (Dict): Config related to `num_channels`.
    """
    dynamic_gn = DynamicGroupNorm(
        num_channels_cfg=in_channels_cfg,
        num_channels=module.num_channels,
        num_groups=module.num_groups,
        eps=module.eps,
        affine=module.affine)
    return dynamic_gn


DEFAULT_MODULE_CONVERTERS: Dict[Callable, Callable] = {
    nn.Conv2d: dynamic_conv2d_converter,
    nn.Linear: dynamic_linear_converter,
    nn.BatchNorm1d: dynamic_bn_converter,
    nn.BatchNorm2d: dynamic_bn_converter,
    nn.BatchNorm3d: dynamic_bn_converter,
    nn.InstanceNorm1d: dynamic_in_converter,
    nn.InstanceNorm2d: dynamic_in_converter,
    nn.InstanceNorm3d: dynamic_in_converter,
    nn.GroupNorm: dynamic_gn_converter
}



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from mmrazor.registry import MODELS


@MODELS.register_module()
class L1Loss(nn.Module):
    """Calculate the one-norm loss between the two inputs.

    Args:
        loss_weight (float): Weight of loss. Defaults to 1.0.
        size_average (bool, optional): Deprecated (see :attr:`reduction`). By
            default, the losses are averaged over each loss element in the
            batch. Note that for some losses, there multiple elements per
            sample. If the field :attr:`size_average` is set to ``False``, the
            losses are instead summed for each minibatch. Ignored when reduce
            is ``False``. Defaults to True.
        reduce (bool, optional): Deprecated (see :attr:`reduction`). By
            default, the losses are averaged or summed over observations for
            each minibatch depending on :attr:`size_average`. When
            :attr:`reduce` is ``False``, returns a loss per batch element
            instead and ignores :attr:`size_average`. Defaults to True.
        reduction (string, optional): Specifies the reduction to apply to the
            output: ``'none'`` | ``'mean'`` | ``'sum'``. ``'none'``: no
            reduction will be applied, ``'mean'``: the sum of the output will
            be divided by the number of elements in the output, ``'sum'``: the
            output will be summed. Note: :attr:`size_average` and
            :attr: `reduce` are in the process of being deprecated, and in the
            meantime, specifying either of those two args will override
            :attr:`reduction`. Defaults to mean.
    """

    def __init__(
        self,
        loss_weight: float = 1.0,
        size_average: Optional[bool] = None,
        reduce: Optional[bool] = None,
        reduction: str = 'mean',
    ) -> None:
        super().__init__()
        self.loss_weight = loss_weight
        self.size_average = size_average
        self.reduce = reduce

        accept_reduction = {'none', 'batchmean', 'sum', 'mean'}
        assert reduction in accept_reduction, \
            f'KLDivergence supports reduction {accept_reduction}, ' \
            f'but gets {reduction}.'
        self.reduction = reduction

    def forward(
        self,
        s_feature: torch.Tensor,
        t_feature: torch.Tensor,
    ) -> torch.Tensor:
        """Forward computation.

        Args:
            s_feature (torch.Tensor): The student model feature with
                shape (N, C, H, W) or shape (N, C).
            t_feature (torch.Tensor): The teacher model feature with
                shape (N, C, H, W) or shape (N, C).
        """
        loss = F.l1_loss(s_feature, t_feature, self.size_average, self.reduce,
                         self.reduction)
        return self.loss_weight * loss



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Union

import torch
import torch.nn as nn


def get_module_device(module: nn.Module) -> torch.device:
    """Get the device of a module.

    Args:
        module (nn.Module): A module contains the parameters.
    """
    try:
        next(module.parameters())
    except StopIteration as e:
        raise ValueError('The input module should contain parameters.') from e

    if next(module.parameters()).is_cuda:
        return next(module.parameters()).get_device()

    return torch.device('cpu')


def set_requires_grad(nets: Union[nn.Module, List[nn.Module]],
                      requires_grad: bool = False) -> None:
    """Set requires_grad for all the networks.

    Args:
        nets (nn.Module | list[nn.Module]): A list of networks or a single
            network.
        requires_grad (bool): Whether the networks require gradients or not
    """
    if not isinstance(nets, list):
        nets = [nets]
    for net in nets:
        if net is not None:
            for param in net.parameters():
                param.requires_grad = requires_grad



# Copyright (c) OpenMMLab. All rights reserved.
from .channel_mutator import (ChannelMutator, OneShotChannelMutator,
                              SlimmableChannelMutator)
from .module_mutator import (DiffModuleMutator, ModuleMutator,
                             OneShotModuleMutator)

__all__ = [
    'OneShotModuleMutator', 'DiffModuleMutator', 'ModuleMutator',
    'ChannelMutator', 'OneShotChannelMutator', 'SlimmableChannelMutator'
]



# Copyright (c) OpenMMLab. All rights reserved.
from abc import abstractmethod
from functools import partial
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from mmrazor.registry import MODELS
from ..base_mutable import CHOICE_TYPE, CHOSEN_TYPE
from .mutable_module import MutableModule

PartialType = Callable[[Any, Optional[nn.Parameter]], Any]


class DiffMutableModule(MutableModule[CHOICE_TYPE, CHOSEN_TYPE]):
    """Base class for differentiable mutables.

    Args:
        module_kwargs (dict[str, dict], optional): Module initialization named
            arguments. Defaults to None.
        alias (str, optional): alias of the `MUTABLE`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.

    Note:
        :meth:`forward_all` is called when calculating FLOPs.
    """

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)

    def forward(self,
                x: Any,
                arch_param: Optional[nn.Parameter] = None) -> Any:
        """Calls either :func:`forward_fixed` or :func:`forward_choice`
        depending on whether :func:`is_fixed` is ``True``.

        To reduce the coupling between `Mutable` and `Mutator`, the
        `arch_param` is generated by the `Mutator` and is passed to the
        forward function as an argument.

        Note:
            :meth:`forward_fixed` is called when in `fixed` mode.
            :meth:`forward_arch_param` is called when in `unfixed` mode.

        Args:
            x (Any): input data for forward computation.
            arch_param (nn.Parameter, optional): the architecture parameters
                for ``DiffMutableModule``.
        """
        if self.is_fixed:
            return self.forward_fixed(x)
        else:
            return self.forward_arch_param(x, arch_param=arch_param)

    def compute_arch_probs(self, arch_param: nn.Parameter) -> Tensor:
        """compute chosen probs according to architecture params."""
        return F.softmax(arch_param, -1)

    @abstractmethod
    def forward_fixed(self, x: Any) -> Any:
        """Forward when the mutable is fixed.

        All subclasses must implement this method.
        """

    @abstractmethod
    def forward_all(self, x: Any) -> Any:
        """Forward all choices."""

    @abstractmethod
    def forward_arch_param(self,
                           x: Any,
                           arch_param: Optional[nn.Parameter] = None) -> Any:
        """Forward when the mutable is not fixed.

        All subclasses must implement this method.
        """

    def set_forward_args(self, arch_param: nn.Parameter) -> None:
        """Interface for modifying the arch_param using partial."""
        forward_with_default_args: PartialType = \
            partial(self.forward, arch_param=arch_param)
        setattr(self, 'forward', forward_with_default_args)


@MODELS.register_module()
class DiffMutableOP(DiffMutableModule[str, str]):
    """A type of ``MUTABLES`` for differentiable architecture search, such as
    DARTS. Search the best module by learnable parameters `arch_param`.

    Args:
        candidates (dict[str, dict]): the configs for the candidate
            operations.
        module_kwargs (dict[str, dict], optional): Module initialization named
            arguments. Defaults to None.
        alias (str, optional): alias of the `MUTABLE`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 5 initializer including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(
        self,
        candidates: Dict[str, Dict],
        module_kwargs: Optional[Dict[str, Dict]] = None,
        alias: Optional[str] = None,
        init_cfg: Optional[Dict] = None,
    ) -> None:
        super().__init__(
            module_kwargs=module_kwargs, alias=alias, init_cfg=init_cfg)
        assert len(candidates) >= 1, \
            f'Number of candidate op must greater than or equal to 1, ' \
            f'but got: {len(candidates)}'

        self._is_fixed = False
        self._candidates = self._build_ops(candidates, self.module_kwargs)

    @staticmethod
    def _build_ops(candidates: Dict[str, Dict],
                   module_kwargs: Optional[Dict[str, Dict]]) -> nn.ModuleDict:
        """Build candidate operations based on candidates configures.

        Args:
            candidates (dict[str, dict]): the configs for the candidate
                operations.
            module_kwargs (dict[str, dict], optional): Module initialization
                named arguments.

        Returns:
            ModuleDict (dict[str, Any], optional):  the key of ``ops`` is
                the name of each choice in configs and the value of ``ops``
                is the corresponding candidate operation.
        """
        ops = nn.ModuleDict()
        for name, op_cfg in candidates.items():
            assert name not in ops
            if module_kwargs is not None:
                op_cfg.update(module_kwargs)
            ops[name] = MODELS.build(op_cfg)
        return ops

    def forward_fixed(self, x: Any) -> Tensor:
        """Forward when the mutable is in `fixed` mode.

        Args:
            x (Any): x could be a Torch.tensor or a tuple of
                Torch.tensor, containing input data for forward computation.

        Returns:
            Tensor: the result of forward the fixed operation.
        """
        return sum(self._candidates[choice](x) for choice in self._chosen)

    def forward_arch_param(self,
                           x: Any,
                           arch_param: Optional[nn.Parameter] = None
                           ) -> Tensor:
        """Forward with architecture parameters.

        Args:
            x (Any): x could be a Torch.tensor or a tuple of
                Torch.tensor, containing input data for forward computation.
            arch_param (str, optional): architecture parameters for
                `DiffMutableModule`


        Returns:
            Tensor: the result of forward with ``arch_param``.
        """
        if arch_param is None:
            return self.forward_all(x)
        else:
            # compute the probs of choice
            probs = self.compute_arch_probs(arch_param=arch_param)

            # forward based on probs
            outputs = list()
            for prob, module in zip(probs, self._candidates.values()):
                if prob > 0.:
                    outputs.append(prob * module(x))

            return sum(outputs)

    def forward_all(self, x: Any) -> Tensor:
        """Forward all choices. Used to calculate FLOPs.

        Args:
            x (Any): x could be a Torch.tensor or a tuple of
                Torch.tensor, containing input data for forward computation.

        Returns:
            Tensor: the result of forward all of the ``choice`` operation.
        """
        outputs = list()
        for op in self._candidates.values():
            outputs.append(op(x))
        return sum(outputs)

    def fix_chosen(self, chosen: Union[str, List[str]]) -> None:
        """Fix mutable with `choice`. This operation would convert `unfixed`
        mode to `fixed` mode. The :attr:`is_fixed` will be set to True and only
        the selected operations can be retained.

        Args:
            chosen (str): the chosen key in ``MUTABLE``.
                Defaults to None.
        """
        if self.is_fixed:
            raise AttributeError(
                'The mode of current MUTABLE is `fixed`. '
                'Please do not call `fix_chosen` function again.')

        if isinstance(chosen, str):
            chosen = [chosen]

        for c in self.choices:
            if c not in chosen:
                self._candidates.pop(c)

        self._chosen = chosen
        self.is_fixed = True

    def sample_choice(self, arch_param):
        """Sample choice based on arch_parameters."""
        return self.choices[torch.argmax(arch_param).item()]

    def dump_chosen(self):
        """Dump current choice."""
        assert self.current_choice is not None
        return self.current_choice

    @property
    def choices(self) -> List[str]:
        """list: all choices. """
        return list(self._candidates.keys())


@MODELS.register_module()
class DiffChoiceRoute(DiffMutableModule[str, List[str]]):
    """A type of ``MUTABLES`` for Neural Architecture Search, which can select
    inputs from different edges in a differentiable or non-differentiable way.
    It is commonly used in DARTS.

    Args:
        edges (nn.ModuleDict): the key of `edges` is the name of different
            edges. The value of `edges` can be :class:`nn.Module` or
            :class:`DiffMutableModule`.
        with_arch_param (bool): whether forward with arch_param. When set to
            `True`, a differentiable way is adopted. When set to `False`,
            a non-differentiable way is adopted.
        alias (str, optional): alias of the `DiffChoiceRoute`.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 6 initializers including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.

    Examples:
        >>> import torch
        >>> import torch.nn as nn
        >>> edges_dict=nn.ModuleDict()
        >>> edges_dict.add_module('first_edge', nn.Conv2d(32, 32, 3, 1, 1))
        >>> edges_dict.add_module('second_edge', nn.Conv2d(32, 32, 5, 1, 2))
        >>> edges_dict.add_module('third_edge', nn.MaxPool2d(3, 1, 1))
        >>> edges_dict.add_module('fourth_edge', nn.MaxPool2d(5, 1, 2))
        >>> edges_dict.add_module('fifth_edge', nn.MaxPool2d(7, 1, 3))
        >>> diff_choice_route_cfg = dict(
        ...     type="DiffChoiceRoute",
        ...     edges=edges_dict,
        ...     with_arch_param=True,
        ... )
        >>> arch_param
        Parameter containing:
        tensor([-6.1426e-04,  2.3596e-04,  1.4427e-03,  7.1668e-05,
            -8.9739e-04], requires_grad=True)
        >>> x = [torch.randn(4, 32, 64, 64) for _ in range(5)]
        >>> output=diffchoiceroute.forward_arch_param(x, arch_param)
        >>> output.shape
        torch.Size([4, 32, 64, 64])
    """

    def __init__(
        self,
        edges: nn.ModuleDict,
        num_chsoen: int = 2,
        with_arch_param: bool = False,
        alias: Optional[str] = None,
        init_cfg: Optional[Dict] = None,
    ) -> None:
        super().__init__(alias=alias, init_cfg=init_cfg)
        assert len(edges) >= 1, \
            f'Number of edges must greater than or equal to 1, ' \
            f'but got: {len(edges)}'

        self._with_arch_param = with_arch_param
        self._is_fixed = False
        self._candidates: nn.ModuleDict = edges
        self.num_chosen = num_chsoen

    def forward_fixed(self, inputs: Union[List, Tuple]) -> Tensor:
        """Forward when the mutable is in `fixed` mode.

        Args:
            inputs (Union[List[Any], Tuple[Any]]): inputs could be a list or
                a tuple of Torch.tensor, containing input data for
                forward computation.

        Returns:
            Tensor: the result of forward the fixed operation.
        """
        assert self._chosen is not None, \
            'Please call fix_chosen before calling `forward_fixed`.'

        outputs = list()
        for choice, x in zip(self._unfixed_choices, inputs):
            if choice in self._chosen:
                outputs.append(self._candidates[choice](x))
        return sum(outputs)

    def forward_arch_param(
            self,
            x: Union[List[Any], Tuple[Any]],
            arch_param: Optional[nn.Parameter] = None) -> Tensor:
        """Forward with architecture parameters.

        Args:
            x (list[Any] | tuple[Any]]): x could be a list or a tuple
                of Torch.tensor, containing input data for forward selection.
            arch_param (nn.Parameter): architecture parameters for
                for ``DiffMutableModule``.

        Returns:
            Tensor: the result of forward with ``arch_param``.
        """
        assert len(x) == len(self._candidates), \
            f'Length of `edges` {len(self._candidates)} should be ' \
            f'same as the length of inputs {len(x)}.'

        if self._with_arch_param:
            probs = self.compute_arch_probs(arch_param=arch_param)

            outputs = list()
            for prob, module, input in zip(probs, self._candidates.values(),
                                           x):
                if prob > 0:
                    # prob may equal to 0 in gumbel softmax.
                    outputs.append(prob * module(input))

            return sum(outputs)
        else:
            return self.forward_all(x)

    def forward_all(self, x: Any) -> Tensor:
        """Forward all choices.

        Args:
            x (Any): x could be a Torch.tensor or a tuple of
                Torch.tensor, containing input data for forward computation.

        Returns:
            Tensor: the result of forward all of the ``choice`` operation.
        """
        assert len(x) == len(self._candidates), \
            f'Lenght of edges {len(self._candidates)} should be same as ' \
            f'the length of inputs {len(x)}.'

        outputs = list()
        for op, input in zip(self._candidates.values(), x):
            outputs.append(op(input))

        return sum(outputs)

    def fix_chosen(self, chosen: List[str]) -> None:
        """Fix mutable with `choice`. This operation would convert to `fixed`
        mode. The :attr:`is_fixed` will be set to True and only the selected
        operations can be retained.

        Args:
            chosen (list(str)): the chosen key in ``MUTABLE``.
        """
        self._unfixed_choices = self.choices

        if self.is_fixed:
            raise AttributeError(
                'The mode of current MUTABLE is `fixed`. '
                'Please do not call `fix_chosen` function again.')

        for c in self.choices:
            if c not in chosen:
                self._candidates.pop(c)

        self._chosen = chosen
        self.is_fixed = True

    @property
    def choices(self) -> List[CHOSEN_TYPE]:
        """list: all choices. """
        return list(self._candidates.keys())

    def dump_chosen(self):
        """dump current choice."""
        assert self.current_choice is not None
        return self.current_choice

    def sample_choice(self, arch_param):
        """sample choice based on `arch_param`."""
        sort_idx = torch.argsort(-arch_param).cpu().numpy().tolist()
        choice_idx = sort_idx[:self.num_chosen]
        choice = [self.choices[i] for i in choice_idx]
        return choice


@MODELS.register_module()
class GumbelChoiceRoute(DiffChoiceRoute):
    """A type of ``MUTABLES`` for Neural Architecture Search using Gumbel-Max
    trick, which can select inputs from different edges in a differentiable or
    non-differentiable way. It is commonly used in DARTS.

    Args:
        edges (nn.ModuleDict): the key of `edges` is the name of different
            edges. The value of `edges` can be :class:`nn.Module` or
            :class:`DiffMutableModule`.
        tau (float): non-negative scalar temperature in gumbel softmax.
        hard (bool): if `True`, the returned samples will be discretized as
            one-hot vectors, but will be differentiated as if it is the soft
            sample in autograd. Defaults to `True`.
        with_arch_param (bool): whether forward with arch_param. When set to
            `True`, a differentiable way is adopted. When set to `False`,
            a non-differentiable way is adopted.
        init_cfg (dict, optional): initialization configuration dict for
            ``BaseModule``. OpenMMLab has implement 6 initializers including
            `Constant`, `Xavier`, `Normal`, `Uniform`, `Kaiming`,
            and `Pretrained`.
    """

    def __init__(
        self,
        edges: nn.ModuleDict,
        tau: float = 1.0,
        hard: bool = True,
        with_arch_param: bool = False,
        alias: Optional[str] = None,
        init_cfg: Optional[Dict] = None,
    ) -> None:
        super().__init__(
            edges=edges,
            with_arch_param=with_arch_param,
            alias=alias,
            init_cfg=init_cfg)
        self.tau = tau
        self.hard = hard

    def compute_arch_probs(self, arch_param: nn.Parameter) -> Tensor:
        """Compute chosen probs by Gumbel-Max trick."""
        return F.gumbel_softmax(
            arch_param, tau=self.tau, hard=self.hard, dim=-1)

    def set_temperature(self, tau: float) -> None:
        """Set temperature of gumbel softmax."""
        self.tau = tau



# Copyright (c) OpenMMLab. All rights reserved.
import functools
from types import FunctionType, ModuleType
from typing import Callable, List, Optional

from mmengine.utils import import_modules_from_strings
from torch import nn

from mmrazor.registry import TASK_UTILS
from .base_recorder import BaseRecorder


@TASK_UTILS.register_module()
class MethodOutputsRecorder(BaseRecorder):
    """Recorder for intermediate results which are ``MethodType``'s outputs.

    Note:
        Different from ``FunctionType``, ``MethodType`` is the type of methods
        of class instances.

    Examples:
        >>> # Below code in toy_module.py
        >>> import random
        >>> class Toy():
        ...     def toy_func(self):
        ...         return random.randint(0, 1000)
        ...     def toy_list_func(self):
        ...         return [random.randint(0, 1000) for _ in range(3)]

        >>> # Below code in main.py
        >>> # Now, we want to get teacher's outputs by recorder.

        >>> from toy_module import Toy
        >>> toy = Toy()
        >>> r1 = MethodOutputsRecorder('toy_module.Toy.toy_func')
        >>> r1.initialize()
        >>> with r1:
        ...     output_teacher1 = toy.toy_func()
        ...     output_teacher2 = toy.toy_func()
        ...     output_teacher3 = toy.toy_func()

        >>> r1.data_buffer
        [33, 41, 12]
        >>> r1.get_record_data(record_idx=2)
        12
        >>> output_teacher1==33 and output_teacher2==41 and output_teacher3==41
        True

        >>> r2 = MethodOutputsRecorder('toy_module.Toy.toy_list_func'
        >>> r2.initialize()
        >>> with r2:
        ...     output_teacher1 = toy.toy_list_func()
        ...     output_teacher2 = toy.toy_list_func()
        ...     output_teacher3 = toy.toy_list_func()

        >>> r2.data_buffer
        [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
        >>> r2.get_record_data(record_idx=2, data_idx=2)
        9
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._check_valid_source(self.source)

        # import the function corrosponding module
        try:
            mod: ModuleType = import_modules_from_strings(self.module_string)
        except ImportError:
            raise ImportError(
                f'{self.module_string} is not imported correctly.')

        assert hasattr(mod, self.cls_name),  \
            f'{self.cls_name} is not in {self.module_string}.'

        imported_cls: type = getattr(mod, self.cls_name)
        if not isinstance(imported_cls, type):
            raise TypeError(f'{self.cls_name} should be a type '
                            f'instance, but got {type(imported_cls)}')
        self.imported_class = imported_cls

        assert hasattr(imported_cls, self.method_name), \
            f'{self.method_name} is not in {self.cls_name}.'

        origin_method = getattr(imported_cls, self.method_name)
        if not isinstance(origin_method, FunctionType):
            raise TypeError(f'{self.method_name} should be a FunctionType '
                            f'instance, but got {type(origin_method)}')
        self.origin_method = origin_method

    @staticmethod
    def _check_valid_source(source: str) -> None:
        """Check if the `source` is valid."""
        if not isinstance(source, str):
            raise TypeError(f'source should be a str '
                            f'instance, but got {type(source)}')

        assert len(source.split('.')) > 2, \
            'source must have at least two `.`'

    @property
    def method_name(self):
        """Get the method name according to `method_path`."""
        return self.source.split('.')[-1]

    @property
    def cls_name(self):
        """Get the class name corresponding to this method according to
        `method_path`."""
        return self.source.split('.')[-2]

    @property
    def module_string(self):
        """Get the module name according to `method_path`."""
        return '.'.join(self.source.split('.')[:-2])

    def prepare_from_model(self, model: Optional[nn.Module] = None) -> None:
        """Wrapper the origin source methods.

        The ``model`` is useless in this recorder, just to be consistent with
        other recorders.
        """
        pass

    def method_record_wrapper(self, orgin_method: Callable,
                              data_buffer: List) -> Callable:
        """Save the method's outputs.

        Args:
            origin_method (MethodType): The method whose outputs need to be
                recorded.
            buffer_key (str): The key of the method's outputs saved in
                ``data_buffer``.
        """

        @functools.wraps(orgin_method)
        def wrap_method(*args, **kwargs):
            outputs = orgin_method(*args, **kwargs)
            # assume a func execute N times, there will be N outputs need to
            # save.
            data_buffer.append(outputs)
            return outputs

        return wrap_method

    def __enter__(self):
        """Enter the context manager."""
        super().__enter__()

        imported_cls = self.imported_class
        origin_method = self.origin_method
        # add record wrapper to origin method.
        record_method = self.method_record_wrapper(origin_method,
                                                   self.data_buffer)

        # rewrite the origin method.
        setattr(imported_cls, self.method_name, record_method)

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager."""
        super().__exit__(exc_type, exc_value, traceback)

        imported_cls = self.imported_class
        origin_method = self.origin_method

        # restore the origin method
        setattr(imported_cls, self.method_name, origin_method)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict

import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor

from mmrazor.models.mutables.base_mutable import BaseMutable
from .dynamic_mixins import DynamicLinearMixin


class DynamicLinear(nn.Linear, DynamicLinearMixin):
    """Dynamic Linear OP.

    Note:
        Arguments for ``__init__`` of ``DynamicLinear`` is totally same as
        :obj:`torch.nn.Linear`.

    Attributes:
        mutable_in_features (BaseMutable, optional): Mutable for controlling
            ``in_features``.
        mutable_out_features (BaseMutable, optional): Mutable for controlling
            ``out_features``.
    """
    accepted_mutable_attrs = {'in_features', 'out_features'}

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)

        self.mutable_attrs: Dict[str, BaseMutable] = nn.ModuleDict()

    @property
    def static_op_factory(self):
        return nn.Linear

    @classmethod
    def convert_from(cls, module):
        """Convert a nn.Linear module to a DynamicLinear.

        Args:
            module (:obj:`torch.nn.Linear`): The original Linear module.
        """
        dynamic_linear = cls(
            in_features=module.in_features,
            out_features=module.out_features,
            bias=True if module.bias is not None else False)
        return dynamic_linear

    def forward(self, input: Tensor) -> Tensor:
        """Forward of dynamic linear OP."""
        weight, bias = self.get_dynamic_params()

        return F.linear(input, weight, bias)



# Copyright (c) OpenMMLab. All rights reserved.
from mmcv.cnn import ConvModule

from mmrazor.registry import MODELS
from .base import BaseOP


@MODELS.register_module()
class Identity(BaseOP):
    """Base class for searchable operations.

    Args:
        conv_cfg (dict, optional): Config dict for convolution layer.
            Default: None, which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer.
            Default: dict(type='BN').
        act_cfg (dict): Config dict for activation layer.
            Default: None.
    """

    def __init__(self,
                 conv_cfg=None,
                 norm_cfg=dict(type='BN'),
                 act_cfg=None,
                 **kwargs):
        super(Identity, self).__init__(**kwargs)

        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg

        if self.stride != 1 or self.in_channels != self.out_channels:
            self.downsample = ConvModule(
                self.in_channels,
                self.out_channels,
                kernel_size=1,
                stride=self.stride,
                padding=0,
                conv_cfg=self.conv_cfg,
                norm_cfg=self.norm_cfg,
                act_cfg=self.act_cfg)
        else:
            self.downsample = None

    def forward(self, x):
        if self.downsample is not None:
            x = self.downsample(x)
        return x



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Union

from mmengine.model import BaseModel

from mmrazor.registry import MODELS
from ....distillers import OFDDistiller
from .single_teacher_distill import SingleTeacherDistill


@MODELS.register_module()
class OverhaulFeatureDistillation(SingleTeacherDistill):
    """`A Comprehensive Overhaul of Feature Distillation`
    https://sites.google.com/view/byeongho-heo/overhaul.

    Inherited from ``SingleTeacherDistill``.


    Args:
        distiller (dict): The config dict for built distiller. Must be a
            ``OFDDistiller``.
        teacher (dict | BaseModel): The config dict for teacher model or built
            teacher model.
        teacher_ckpt (str): The path of teacher's checkpoint. Defaults to None.
        teacher_trainable (bool): Whether the teacher is trainable. Defaults
            to False.
        teacher_norm_eval (bool): Whether to set teacher's norm layers to eval
            mode, namely, freeze running stats (mean and var). Note: Effect on
            Batch Norm and its variants only. Defaults to True.
        student_trainable (bool): Whether the student is trainable. Defaults
            to True.
        calculate_student_loss (bool): Whether to calculate student loss
            (original task loss) to update student model. Defaults to True.
    """

    def __init__(self,
                 distiller: dict,
                 teacher: Union[BaseModel, Dict],
                 teacher_ckpt: Optional[str] = None,
                 teacher_trainable: bool = False,
                 teacher_norm_eval: bool = True,
                 student_trainable: bool = True,
                 calculate_student_loss: bool = True,
                 **kwargs) -> None:
        super().__init__(distiller, teacher, teacher_ckpt, teacher_trainable,
                         teacher_norm_eval, student_trainable,
                         calculate_student_loss, **kwargs)

        assert isinstance(self.distiller, OFDDistiller), (
            'distiller of `OverhaulFeatureDistillation` expects `OFDDistiller`'
            f', but get {type(self.distiller)}')

        self.distiller.init_ofd_connectors(self.teacher)



# Copyright (c) OpenMMLab. All rights reserved.
from .delivery_manager import DistillDeliveryManager
from .function_outputs_delivery import FunctionOutputsDelivery
from .method_outputs_delivery import MethodOutputsDelivery

__all__ = [
    'FunctionOutputsDelivery', 'MethodOutputsDelivery',
    'DistillDeliveryManager'
]



# Copyright (c) OpenMMLab. All rights reserved.
from .base_distiller import BaseDistiller
from .byot_distiller import BYOTDistiller
from .configurable_distiller import ConfigurableDistiller
from .ofd_distiller import OFDDistiller

__all__ = [
    'ConfigurableDistiller', 'BaseDistiller', 'BYOTDistiller', 'OFDDistiller'
]



# Copyright (c) OpenMMLab. All rights reserved.
import os
from pathlib import Path

import requests
import yaml

MMRAZOR_ROOT = Path(__file__).absolute().parents[1]


class TestMetafiles:

    def get_metafiles(self, code_path):
        """
        Function: get the metafile of all configs from model-index.yml
        """
        metafile = os.path.join(code_path, 'model-index.yml')
        with open(metafile, 'r') as f:
            meta = yaml.safe_load(f)
        return meta['Import']

    def test_metafiles(self):
        metafiles = self.get_metafiles(MMRAZOR_ROOT)
        for mf in metafiles:
            metafile = os.path.abspath(os.path.join(MMRAZOR_ROOT, mf))
            with open(metafile, 'r') as f:
                meta = yaml.safe_load(f)
            for model in meta['Models']:
                # 1. weights url check
                r = requests.head(model['Weights'], timeout=4)
                assert r.status_code != 404, \
                    f"can't connect url {model['Weights']} in " \
                    f'metafile {metafile}'

                # 2. config check
                dir_path = os.path.abspath(os.path.join(metafile, '../'))
                # list all files which are in the same directory of
                # current metafile
                config_files = os.listdir(dir_path)

                if isinstance(model['Config'], list):
                    # TODO: 3. log error
                    continue

                assert (model['Config'].split('/')[-1] in config_files), \
                    f"config error in {metafile} model {model['Name']}"

                # 4. name check
                # erase '.py'
                correct_name = model['Config'].split('/')[-1][:-3]
                assert model['Name'] == correct_name, \
                    f'name error in {metafile}, correct name should ' \
                    f'be {correct_name}'



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from pathlib import Path
from typing import Dict, List, Optional, Union

import torch.nn as nn
from torch.nn import Module
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.models.architectures.dynamic_ops import DynamicBatchNorm
from mmrazor.models.mutables import SlimmableMutableChannel
from mmrazor.registry import MODELS
from ...task_modules import PathList
from ..utils import switchable_bn_converter
from .channel_mutator import ChannelMutator

NONPASS_MODULES = (nn.Conv2d, nn.Linear)
PASS_MODULES = (_BatchNorm, )

VALID_PATH_TYPE = Union[str, Path]


@MODELS.register_module()
class SlimmableChannelMutator(ChannelMutator):
    """Slimmable channel mutable based channel mutator.

    Args:
        channel_cfgs (list[Dict]): A list of candidate channel configs.
        mutable_cfg (dict): The config for the channel mutable.
        skip_prefixes (List[str] | Optional): The module whose name start with
            a string in skip_prefixes will not be pruned.
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(self,
                 channel_cfgs: Dict,
                 mutable_cfg: Dict,
                 tracer_cfg: Dict,
                 skip_prefixes: Optional[List[str]] = None,
                 init_cfg: Optional[Dict] = None):
        super(SlimmableChannelMutator, self).__init__(
            mutable_cfg=mutable_cfg,
            tracer_cfg=tracer_cfg,
            skip_prefixes=skip_prefixes,
            init_cfg=init_cfg)

        self.channel_cfgs = channel_cfgs

    def prepare_from_supernet(self, supernet: Module) -> None:
        """Do some necessary preparations with supernet.

        Note:
            Different from `ChannelMutator`, we only support Case 1 in
            `ChannelMutator`. The input supernet should be made up of original
            nn.Module. And we replace the conv/linear/bn modules in the input
            supernet with dynamic ops first. Then we trace the topology of
            the supernet to get the `concat_parent_mutables` of a certain
            mutable, if the input of a module is a concatenation of several
            modules' outputs. Then we convert the ``DynamicBatchNorm`` in
            supernet with ``SwitchableBatchNorm2d``, and set the candidate
            channel numbers to the corresponding `SlimmableChannelMutable`.
            Finally, we establish the relationship between the current nodes
            and their parents.

        Args:
            supernet (:obj:`torch.nn.Module`): The supernet to be searched
                in your algorithm.
        """
        self.convert_dynamic_module(supernet, self.module_converters)

        module_path_list: PathList = self.tracer.trace(supernet)

        self.convert_switchable_bn(supernet)
        self.set_candidate_choices(supernet)

        # The mapping from a module name to the module
        self._name2module = dict(supernet.named_modules())
        self.add_link(module_path_list)
        self.bind_mutable_name(supernet)

    def set_candidate_choices(self, supernet):
        """Set the ``candidate_choices`` of each ``SlimmableChannelMutable``.

        Notes:
            Different from other ``OneShotChannelMutable``,
            ``candidate_choices`` is optional when instantiating a
            ``SlimmableChannelMutable``
        """
        for name, module in supernet.named_modules():
            if isinstance(module, SlimmableMutableChannel):
                candidate_choices = self.channel_cfgs[name]['current_choice']
                module.candidate_choices = candidate_choices

    def convert_switchable_bn(self, supernet):
        """Replace ``DynamicBatchNorm`` in supernet with
        ``SwitchableBatchNorm2d``.

        Args:
            supernet (:obj:`torch.nn.Module`): The architecture to be converted
                in your algorithm.
        """

        def traverse(module, prefix):
            for name, child in module.named_children():
                module_name = prefix + name
                if isinstance(child, DynamicBatchNorm):
                    mutable_cfg = copy.deepcopy(self.mutable_cfg)
                    key = module_name + '.mutable_num_features'
                    candidate_choices = self.channel_cfgs[key][
                        'current_choice']
                    mutable_cfg.update(
                        dict(candidate_choices=candidate_choices))
                    sbn = switchable_bn_converter(child, mutable_cfg,
                                                  mutable_cfg)
                    # TODO
                    # bind twice?
                    sbn.mutable_out.bind_mutable_name(module_name)
                    setattr(module, name, sbn)
                else:
                    traverse(child, module_name + '.')

        traverse(supernet, '')

    def switch_choices(self, idx: int) -> None:
        """Switch the channel config of the supernet according to input `idx`.

        If we train more than one subnet together, we need to switch the
        channel_cfg from one to another during one training iteration.

        Args:
            idx (int): The index of the current subnet.
        """
        for name, module in self.name2module.items():
            if isinstance(module, SlimmableMutableChannel):
                module.current_choice = idx

    def build_search_groups(self, supernet: Module):
        """Build `search_groups`.

        The mutables in the same group should be pruned together.
        """
        pass

    def mutable_class_type(self):
        """One-shot channel mutable class type.

        Returns:
            Type[OneShotMutableModule]: Class type of one-shot mutable.
        """
        return SlimmableMutableChannel



# Copyright (c) OpenMMLab. All rights reserved.
from .autoslim import AutoSlim, AutoSlimDDP
from .darts import Darts, DartsDDP
from .spos import SPOS

__all__ = ['SPOS', 'AutoSlim', 'AutoSlimDDP', 'Darts', 'DartsDDP']



# Copyright (c) OpenMMLab. All rights reserved.
# This file is modified from `mmcls.models.backbones.resnet`

import warnings
from typing import Dict, Tuple

import torch
import torch.nn as nn
import torch.nn.functional as F
from mmcv.cnn import build_conv_layer, build_norm_layer
from mmengine.model import BaseModule
from mmengine.model.weight_init import constant_init

from mmrazor.registry import MODELS


class BasicBlock(nn.Module):
    """BasicBlock for WideResNet. The differences from ResNet are in:
     1. The forward path
     2. The position of residual path
     3. Different downsample

    Args:
        in_channels (int): Input channels of this block.
        out_channels (int): Output channels of this block.
        expansion (int): The ratio of ``out_channels/mid_channels`` where
            ``mid_channels`` is the output channels of conv1. This is a
            reserved argument in BasicBlock and should always be 1. Default: 1.
        stride (int): stride of the block. Default: 1
        stride (int): stride of the block. Default: 1
        dilation (int): dilation of convolution. Default: 1
        downsample (nn.Module, optional): downsample operation on identity
            branch. Default: None.
        droprate (float, optional): droprate of the block. Defaults to 0.
        conv_cfg (dict, optional): dictionary to construct and config conv
            layer. Default: None
        norm_cfg (dict): dictionary to construct and config norm layer.
            Default: dict(type='BN')
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        expansion: int = 1,
        stride: int = 1,
        dilation: int = 1,
        downsample: nn.Module = None,
        droprate: float = 0,
        conv_cfg: Dict = None,
        norm_cfg: Dict = dict(type='BN')
    ) -> None:  # noqa: E125
        super(BasicBlock, self).__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.expansion = expansion
        self.stride = stride
        self.dilation = dilation
        self.droprate = droprate
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg

        self.norm1_name, norm1 = build_norm_layer(
            norm_cfg, in_channels, postfix=1)
        self.norm2_name, norm2 = build_norm_layer(
            norm_cfg, out_channels, postfix=2)

        self.add_module(self.norm1_name, norm1)
        self.relu1 = nn.ReLU(inplace=True)
        self.conv1 = build_conv_layer(
            conv_cfg,
            in_channels,
            out_channels,
            3,
            stride=stride,
            padding=dilation,
            dilation=dilation,
            bias=False)

        self.add_module(self.norm2_name, norm2)
        self.relu2 = nn.ReLU(inplace=True)
        self.conv2 = build_conv_layer(
            conv_cfg,
            out_channels,
            out_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False)
        self.downsample = downsample

    @property
    def norm1(self):
        return getattr(self, self.norm1_name)

    @property
    def norm2(self):
        return getattr(self, self.norm2_name)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """forward func.

        Args:
            x (torch.Tensor): input.

        Returns:
            torch.Tensor: output.
        """

        identity = self.relu1(self.bn1(x))
        out = self.conv1(identity)
        out = self.bn2(out)
        out = self.relu2(out)
        if self.droprate > 0:
            out = F.dropout(out, p=self.droprate, training=self.training)
        out = self.conv2(out)
        if self.downsample:
            out += self.downsample(identity)
        else:
            out += x
        return out


def get_expansion(block: nn.Module,
                  widen_factor: int,
                  expansion: int = None) -> int:
    """Get the expansion of a residual block.
    The block expansion will be obtained by the following order:
    1. If ``expansion`` is given, just return it.
    2. If ``block`` has the attribute ``expansion``, then return
       ``block.expansion``.
    3. If ``block`` is ``BaseBlock``, then return ``widen_factor``.
    3. Return the default value according the the block type:
       4 for ``Bottleneck``.

    Args:
        block (class): The block class.
        widen_factor (int): The given widen factor.
        expansion (int | None): The given expansion ratio.
    Returns:
        int: The expansion of the block.
    """
    if isinstance(expansion, int):
        assert expansion > 0
    elif expansion is None:
        if hasattr(block, 'expansion'):
            expansion = block.expansion
        elif issubclass(block, BasicBlock):
            expansion = widen_factor
        else:
            raise TypeError(f'expansion is not specified for {block.__name__}')
    else:
        raise TypeError('expansion must be an integer or None')

    return expansion


class ResLayer(nn.Sequential):
    """ResLayer to build ResNet style backbone.

    Args:
        block (nn.Module): Residual block used to build ResLayer.
        num_blocks (int): Number of blocks.
        in_channels (int): Input channels of this block.
        out_channels (int): Output channels of this block.
        expansion (int): The expansion for BasicBlock/Bottleneck.
            If not specified, it will firstly be obtained via
            ``block.expansion``. If the block has no attribute "expansion",
            the following default values will be used: 1 for BasicBlock and
            4 for Bottleneck.
        droprate (float, optional): droprate of the layer. Defaults to 0.
        stride (int): stride of the first block. Default: 1.
        conv_cfg (Dict, optional): dictionary to construct and config conv
            layer. Default: None
        norm_cfg (Dict): dictionary to construct and config norm layer.
            Default: dict(type='BN')
    """

    def __init__(self,
                 block: nn.Module,
                 num_blocks: int,
                 in_channels: int,
                 out_channels: int,
                 expansion: int,
                 droprate: float = 0,
                 stride: int = 1,
                 conv_cfg: Dict = None,
                 norm_cfg: Dict = dict(type='BN'),
                 **kwargs):
        self.block = block
        self.droprate = droprate
        self.expansion = expansion

        downsample = None
        if stride != 1 or in_channels != out_channels:
            downsample = build_conv_layer(
                conv_cfg,
                in_channels,
                out_channels,
                kernel_size=1,
                stride=stride,
                bias=False)

        layers = []
        layers.append(
            block(
                in_channels=in_channels,
                out_channels=out_channels,
                expansion=self.expansion,
                stride=stride,
                downsample=downsample,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
                **kwargs))
        in_channels = out_channels
        for _ in range(1, num_blocks):
            layers.append(
                block(
                    in_channels=in_channels,
                    out_channels=out_channels,
                    expansion=self.expansion,
                    stride=1,
                    conv_cfg=conv_cfg,
                    norm_cfg=norm_cfg,
                    **kwargs))
        super(ResLayer, self).__init__(*layers)


@MODELS.register_module()
class WideResNet(BaseModule):
    """WideResNet backbone. Only support 3-stage WideResNet, which is usually
    for tiny images. E.g., CIFAR10 and CIFAR100.

    WRN50 and WRN101 are now officially supported in
    MMClassification. See link below:
    https://github.com/open-mmlab/mmclassification/pull/715

    Please refer to the `paper <https://arxiv.org/abs/1605.07146>`__ for
    details.

    Args:
        depth (int): Network depth, from {10, 16, 22, 28, 40, 50, 101, 152}.
        widen_factor (int):  Width multiplier of mid-channel in blocks.
        in_channels (int): Number of input image channels. Default: 3.
        stem_channels (int): Output channels of the stem layer. Default: 64.
        base_channels (int): Middle channels of the first stage. Default: 64.
        num_stages (int): Stages of the network. Default: 4.
        strides (Sequence[int]): Strides of the first block of each stage.
            Default: ``(1, 2, 2, 2)``.
        dilations (Sequence[int]): Dilation of each stage.
            Default: ``(1, 1, 1, 1)``.
        frozen_stages (int): Stages to be frozen (stop grad and set eval mode).
            -1 means not freezing any parameters. Default: -1.
        conv_cfg (dict | None): The config dict for conv layers. Default: None.
        norm_cfg (dict): The config dict for norm layers.
        norm_eval (bool): Whether to set norm layers to eval mode, namely,
            freeze running stats (mean and var). Note: Effect on Batch Norm
            and its variants only. Default: False.
        zero_init_residual (bool): Whether to use zero init for last norm layer
            in resblocks to let them behave as identity. Default: True.
    """
    arch_settings = {
        10: (BasicBlock, (1, 1, 1)),
        16: (BasicBlock, (2, 2, 2)),
        22: (BasicBlock, (3, 3, 3)),
        28: (BasicBlock, (4, 4, 4)),
        40: (BasicBlock, (6, 6, 6)),
    }

    def __init__(self,
                 depth: int,
                 widen_factor: int = 4,
                 in_channels: int = 3,
                 stem_channels: int = 16,
                 base_channels: int = 16,
                 expansion: int = None,
                 num_stages: int = 3,
                 strides: Tuple[int, ...] = (1, 2, 2),
                 dilations: Tuple[int, ...] = (1, 1, 1),
                 frozen_stages: int = -1,
                 conv_cfg: Dict = None,
                 norm_cfg: Dict = dict(type='BN', requires_grad=True),
                 norm_eval: bool = False,
                 zero_init_residual: bool = False,
                 init_cfg=[
                     dict(type='Kaiming', layer=['Conv2d']),
                     dict(
                         type='Constant',
                         val=1,
                         layer=['_BatchNorm', 'GroupNorm'])
                 ]):
        super(WideResNet, self).__init__(init_cfg)
        if depth > 40:
            """MMClassication now supports WRN-50 and 101 officially.

            Refer to:
            https://github.com/open-mmlab/mmclassification/pull/715/files
            """
            warnings.warn('`WiderResNet` deep than 40 now is deprecated')
        if depth not in self.arch_settings:
            raise KeyError(f'invalid depth {depth} for WideResNet')
        self.depth = depth
        self.widen_factor = widen_factor
        self.stem_channels = stem_channels
        self.base_channels = base_channels
        self.num_stages = num_stages
        self.strides = strides
        self.dilations = dilations
        self.frozen_stages = frozen_stages
        self.conv_cfg = conv_cfg
        self.norm_cfg = norm_cfg
        self.norm_eval = norm_eval
        self.zero_init_residual = zero_init_residual
        self.block, stage_blocks = self.arch_settings[depth]
        self.stage_blocks = stage_blocks[:num_stages]
        self.expansion = get_expansion(self.block, widen_factor, expansion)

        self._make_stem_layer(in_channels, stem_channels)

        self.res_layers = []
        _in_channels = stem_channels
        _out_channels = base_channels * self.expansion
        for i, num_blocks in enumerate(self.stage_blocks):
            stride = strides[i]
            dilation = dilations[i]
            res_layer = self.make_res_layer(
                block=self.block,
                num_blocks=num_blocks,
                in_channels=_in_channels,
                out_channels=_out_channels,
                expansion=self.expansion,
                stride=stride,
                dilation=dilation,
                conv_cfg=conv_cfg,
                norm_cfg=norm_cfg,
            )
            _in_channels = _out_channels
            _out_channels *= 2
            layer_name = f'layer{i + 1}'
            self.add_module(layer_name, res_layer)
            self.res_layers.append(layer_name)

        self._freeze_stages()

        self.feat_dim = res_layer[-1].out_channels

        self.norm1_name, norm1 = build_norm_layer(
            self.norm_cfg, _out_channels // 2, postfix=1)
        self.add_module(self.norm1_name, norm1)
        self.relu = nn.ReLU(inplace=True)

    @property
    def norm1(self):
        return getattr(self, self.norm1_name)

    def make_res_layer(self, **kwargs):
        return ResLayer(**kwargs)

    def _make_stem_layer(self, in_channels, base_channels):
        self.conv1 = build_conv_layer(
            self.conv_cfg,
            in_channels,
            base_channels,
            kernel_size=3,
            stride=1,
            padding=1,
            bias=False)

    def _freeze_stages(self):
        if self.frozen_stages >= 0:
            self.norm1.eval()
            for m in [self.conv1, self.norm1]:
                for param in m.parameters():
                    param.requires_grad = False

        for i in range(1, self.frozen_stages + 1):
            m = getattr(self, f'layer{i}')
            m.eval()
            for param in m.parameters():
                param.requires_grad = False

    def init_weights(self):
        super(WideResNet, self).init_weights()

        if (isinstance(self.init_cfg, dict)
                and self.init_cfg['type'] == 'Pretrained'):
            # Suppress zero_init_residual if use pretrained model.
            return

        if self.zero_init_residual:
            for m in self.modules():
                if isinstance(m, BasicBlock):
                    constant_init(m.norm2, 0)

    def forward(self, x):
        # TODO: return multi-stage features.
        x = self.conv1(x)
        for layer_name in self.res_layers:
            res_layer = getattr(self, layer_name)
            x = res_layer(x)
        x = self.norm1(x)
        x = self.relu(x)
        return tuple([x])



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABC, abstractmethod
from typing import Dict, Optional

from mmengine.model import BaseModule

from ..algorithms.base import LossResults


class BaseDistiller(BaseModule, ABC):
    """Base class for distiller.

    Args:
        init_cfg (dict, optional): Config for distiller. Default to None.
    """

    def __init__(self, init_cfg: Optional[Dict] = None) -> None:
        super().__init__(init_cfg)

    @abstractmethod
    def compute_distill_losses(self) -> LossResults:
        """Compute distill losses automatically."""



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional, Tuple, Union

import torch
from mmcv.cnn import ConvModule

from mmrazor.registry import MODELS
from .base_connector import BaseConnector


@MODELS.register_module()
class ConvModuleConncetor(BaseConnector):
    """Convolution connector that bundles conv/norm/activation layers.

    Args:
        in_channel (int): The input channel of the connector.
        out_channel (int): The output channel of the connector.
        kernel_size (int | tuple[int, int]): Size of the convolving kernel.
            Same as that in ``nn._ConvNd``.
        stride (int | tuple[int, int]): Stride of the convolution.
            Same as that in ``nn._ConvNd``.
        padding (int | tuple[int, int]): Zero-padding added to both sides of
            the input. Same as that in ``nn._ConvNd``.
        dilation (int | tuple[int, int]): Spacing between kernel elements.
            Same as that in ``nn._ConvNd``.
        groups (int): Number of blocked connections from input channels to
            output channels. Same as that in ``nn._ConvNd``.
        bias (bool | str): If specified as `auto`, it will be decided by the
            norm_cfg. Bias will be set as True if `norm_cfg` is None, otherwise
            False. Default: "auto".
        conv_cfg (dict): Config dict for convolution layer. Default: None,
            which means using conv2d.
        norm_cfg (dict): Config dict for normalization layer. Default: None.
        act_cfg (dict): Config dict for activation layer.
            Default: dict(type='ReLU').
        inplace (bool): Whether to use inplace mode for activation.
            Default: True.
        with_spectral_norm (bool): Whether use spectral norm in conv module.
            Default: False.
        padding_mode (str): If the `padding_mode` has not been supported by
            current `Conv2d` in PyTorch, we will use our own padding layer
            instead. Currently, we support ['zeros', 'circular'] with official
            implementation and ['reflect'] with our own implementation.
            Default: 'zeros'.
        order (tuple[str]): The order of conv/norm/activation layers. It is a
            sequence of "conv", "norm" and "act". Common examples are
            ("conv", "norm", "act") and ("act", "conv", "norm").
            Default: ('conv', 'norm', 'act').
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(
        self,
        in_channel: int,
        out_channel: int,
        kernel_size: Union[int, Tuple[int, int]] = 1,
        stride: Union[int, Tuple[int, int]] = 1,
        padding: Union[int, Tuple[int, int]] = 0,
        dilation: Union[int, Tuple[int, int]] = 1,
        groups: int = 1,
        bias: Union[str, bool] = 'auto',
        conv_cfg: Optional[Dict] = None,
        norm_cfg: Optional[Dict] = None,
        act_cfg: Dict = dict(type='ReLU'),
        inplace: bool = True,
        with_spectral_norm: bool = False,
        padding_mode: str = 'zeros',
        order: tuple = ('conv', 'norm', 'act'),
        init_cfg: Optional[Dict] = None,
    ) -> None:
        super().__init__(init_cfg)
        self.conv_module = ConvModule(in_channel, out_channel, kernel_size,
                                      stride, padding, dilation, groups, bias,
                                      conv_cfg, norm_cfg, act_cfg, inplace,
                                      with_spectral_norm, padding_mode, order)

    def forward_train(self, feature: torch.Tensor) -> torch.Tensor:
        """Forward computation.

        Args:
            feature (torch.Tensor): Input feature.
        """
        for layer in self.conv_module.order:
            if layer == 'conv':
                if self.conv_module.with_explicit_padding:
                    feature = self.conv_module.padding_layer(feature)
                feature = self.conv_module.conv(feature)
            elif layer == 'norm' and self.conv_module.with_norm:
                feature = self.conv_module.norm(feature)
            elif layer == 'act' and self.conv_module.with_activation:
                feature = self.conv_module.activate(feature)
        return feature



_base_ = [
    'mmcls::_base_/datasets/cifar10_bs16.py',
    'mmcls::_base_/schedules/cifar10_bs128.py',
    'mmcls::_base_/default_runtime.py'
]

train_cfg = dict(by_epoch=True, max_epochs=20, val_interval=1)

model = dict(
    _scope_='mmrazor',
    type='SingleTeacherDistill',
    data_preprocessor=dict(
        type='ImgDataPreprocessor',
        # RGB format normalization parameters
        mean=[123.675, 116.28, 103.53],
        std=[58.395, 57.12, 57.375],
        # convert image from BGR to RGB
        bgr_to_rgb=True),
    architecture=dict(
        cfg_path='mmcls::resnet/resnet18_8xb16_cifar10.py', pretrained=False),
    teacher=dict(
        cfg_path='mmcls::resnet/resnet50_8xb16_cifar10.py', pretrained=True),
    teacher_ckpt=  # noqa: E251
    'https://download.openmmlab.com/mmclassification/v0/resnet/resnet50_b16x8_cifar10_20210528-f54bfad9.pth',  # noqa: E501
    calculate_student_loss=False,
    student_trainable=False,
    distiller=dict(
        type='ConfigurableDistiller',
        student_recorders=dict(
            bb_s4=dict(type='ModuleOutputs',
                       source='backbone.layer4.1.conv2')),
        teacher_recorders=dict(
            bb_s4=dict(type='ModuleOutputs',
                       source='backbone.layer4.2.conv3')),
        distill_losses=dict(
            loss_s4_pretrain=dict(type='L2Loss', loss_weight=1.0)),
        connectors=dict(
            loss_s4_sfeat=dict(
                type='Translator', in_channel=512, out_channel=1024),
            loss_s4_tfeat=dict(
                type='Paraphraser',
                phase='pretrain',
                in_channel=2048,
                out_channel=1024)),
        loss_forward_mappings=dict(
            loss_s4_pretrain=dict(
                s_feature=dict(
                    # it actually is t_feature
                    from_student=False,
                    recorder='bb_s4'),
                t_feature=dict(
                    from_student=False,
                    recorder='bb_s4',
                    connector='loss_s4_tfeat'),
            ))))

find_unused_parameters = True

val_cfg = dict(_delete_=True, type='mmrazor.SingleTeacherDistillValLoop')



_base_ = [
    'mmrazor::_base_/settings/cifar10_darts_supernet.py',
    'mmrazor::_base_/nas_backbones/darts_supernet.py',
    'mmcls::_base_/default_runtime.py',
]

# model
mutator = dict(type='mmrazor.DiffModuleMutator')

model = dict(
    type='mmrazor.Darts',
    architecture=dict(
        type='ImageClassifier',
        backbone=_base_.nas_backbone,
        neck=dict(type='GlobalAveragePooling'),
        head=dict(
            type='LinearClsHead',
            num_classes=10,
            in_channels=256,
            loss=dict(type='CrossEntropyLoss', loss_weight=1.0),
            topk=(1, 5),
            cal_acc=True)),
    mutator=dict(type='mmrazor.DiffModuleMutator'),
    unroll=True)

model_wrapper_cfg = dict(
    type='mmrazor.DartsDDP',
    broadcast_buffers=False,
    find_unused_parameters=False)

# TRAINING
optim_wrapper = dict(
    _delete_=True,
    constructor='mmrazor.SeparateOptimWrapperConstructor',
    architecture=dict(
        optimizer=dict(type='SGD', lr=0.025, momentum=0.9, weight_decay=3e-4),
        clip_grad=dict(max_norm=5, norm_type=2)),
    mutator=dict(optimizer=dict(type='Adam', lr=3e-4, weight_decay=1e-3)))

find_unused_parameter = False



# Copyright (c) OpenMMLab. All rights reserved.
from abc import ABCMeta, abstractmethod
from typing import Dict, Optional, Tuple, Union

import torch
from mmengine.model import BaseModule


class BaseConnector(BaseModule, metaclass=ABCMeta):
    """Base class of connectors.

    Connector is mainly used for distillation, it usually converts the channel
    number of input feature to align features of student and teacher.

    All subclasses should implement the following APIs:

    - ``forward_train()``

    Args:
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(self, init_cfg: Optional[Dict] = None) -> None:
        super().__init__(init_cfg=init_cfg)

    def forward(self, feature: torch.Tensor) -> torch.Tensor:
        """Forward computation.

        Args:
            feature (torch.Tensor): Input feature.
        """
        return self.forward_train(feature)

    @abstractmethod
    def forward_train(
        self, feature: torch.Tensor
    ) -> Union[Tuple[torch.Tensor, ...], torch.Tensor]:
        """Abstract train computation.

        Args:
            feature (torch.Tensor): Input feature.
        """
        pass



# Copyright (c) OpenMMLab. All rights reserved.
from typing import List, Optional

from mmrazor.registry import MODELS
from .configurable_distiller import ConfigurableDistiller


@MODELS.register_module()
class BYOTDistiller(ConfigurableDistiller):
    """``BYOTDistiller`` inherits ``ConfigurableDistiller`` and only modifies
    ``get_record()`` function to ``get_record_with_cidx()``.

    In ``BYOTDistiller``, ``self.teacher_recorder`` records self-teacher data
    which requires detach().
    """

    def get_record(self,
                   recorder: str,
                   from_student: bool,
                   record_idx: int = 0,
                   data_idx: Optional[int] = None,
                   connector: Optional[str] = None,
                   connector_idx: Optional[int] = None) -> List:
        """According to each item in ``record_infos``, get the corresponding
        record in ``recorder_manager``.

        Detach teacher_record.
        """

        if from_student:
            recorder_ = self.student_recorders.get_recorder(recorder)
        else:
            recorder_ = self.teacher_recorders.get_recorder(recorder)
        record_data = recorder_.get_record_data(record_idx, data_idx)

        if connector:
            record_data = self.connectors[connector](record_data)
        if connector_idx is not None:
            record_data = record_data[connector_idx]
        # Detach self-teacher output Tensor from model, assert hook tensor.
        if not from_student:
            record_data = record_data.detach()

        return record_data



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Union

import torch
from mmengine.model import BaseModel
from mmengine.runner import load_checkpoint
from mmengine.structures import BaseDataElement
from torch import nn
from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.models.utils import add_prefix
from mmrazor.registry import MODELS
from ...base import BaseAlgorithm, LossResults


@MODELS.register_module()
class SingleTeacherDistill(BaseAlgorithm):
    """``SingleTeacherDistill`` can be used to develop distill algorithms which
    only use one teacher.

    Args:
        distiller (dict): The config dict for built distiller.
        teacher (dict | BaseModel): The config dict for teacher model or built
            teacher model.
        teacher_ckpt (str): The path of teacher's checkpoint. Defaults to None.
        teacher_trainable (bool): Whether the teacher is trainable. Defaults
            to False.
        teacher_norm_eval (bool): Whether to set teacher's norm layers to eval
            mode, namely, freeze running stats (mean and var). Note: Effect on
            Batch Norm and its variants only. Defaults to True.
        student_trainable (bool): Whether the student is trainable. Defaults
            to True.
        calculate_student_loss (bool): Whether to calculate student loss
            (original task loss) to update student model. Defaults to True.
    """

    def __init__(self,
                 distiller: dict,
                 teacher: Union[BaseModel, Dict],
                 teacher_ckpt: Optional[str] = None,
                 teacher_trainable: bool = False,
                 teacher_norm_eval: bool = True,
                 student_trainable: bool = True,
                 calculate_student_loss: bool = True,
                 **kwargs) -> None:
        super().__init__(**kwargs)

        self.distiller = MODELS.build(distiller)

        if isinstance(teacher, Dict):
            teacher = MODELS.build(teacher)

        if not isinstance(teacher, BaseModel):
            raise TypeError('teacher should be a `dict` or '
                            f'`BaseModel` instance, but got '
                            f'{type(teacher)}')

        self.teacher = teacher
        if teacher_ckpt:
            # avoid loaded parameters be overwritten
            self.teacher.init_weights()
            _ = load_checkpoint(self.teacher, teacher_ckpt)
        self.teacher_trainable = teacher_trainable
        self.teacher_norm_eval = teacher_norm_eval

        # The student model will not calculate gradients and update parameters
        # in some pretraining process.
        self.student_trainable = student_trainable

        # The student loss will not be updated into ``losses`` in some
        # pretraining process.
        self.calculate_student_loss = calculate_student_loss

        # In ``ConfigurableDistller``, the recorder manager is just
        # constructed, but not really initialized yet.
        self.distiller.prepare_from_student(self.student)
        self.distiller.prepare_from_teacher(self.teacher)

    @property
    def student(self) -> nn.Module:
        """Alias for ``architecture``."""
        return self.architecture

    def loss(
        self,
        batch_inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
    ) -> LossResults:
        """Calculate losses from a batch of inputs and data samples."""

        losses = dict()

        # If the `override_data` of a delivery is False, the delivery will
        # record the origin data.
        self.distiller.set_deliveries_override(False)
        if self.teacher_trainable:
            with self.distiller.teacher_recorders, self.distiller.deliveries:
                teacher_losses = self.teacher(
                    batch_inputs, data_samples, mode='loss')

            losses.update(add_prefix(teacher_losses, 'teacher'))
        else:
            with self.distiller.teacher_recorders, self.distiller.deliveries:
                with torch.no_grad():
                    _ = self.teacher(batch_inputs, data_samples, mode='loss')

        # If the `override_data` of a delivery is True, the delivery will
        # override the origin data with the recorded data.
        self.distiller.set_deliveries_override(True)
        # Original task loss will not be used during some pretraining process.
        if self.calculate_student_loss:
            with self.distiller.student_recorders, self.distiller.deliveries:
                student_losses = self.student(
                    batch_inputs, data_samples, mode='loss')
            losses.update(add_prefix(student_losses, 'student'))
        else:
            with self.distiller.student_recorders, self.distiller.deliveries:
                if self.student_trainable:
                    _ = self.student(batch_inputs, data_samples, mode='loss')
                else:
                    with torch.no_grad():
                        _ = self.student(
                            batch_inputs, data_samples, mode='loss')

        # Automatically compute distill losses based on `loss_forward_mappings`
        # The required data already exists in the recorders.
        distill_losses = self.distiller.compute_distill_losses()
        losses.update(add_prefix(distill_losses, 'distill'))

        return losses

    def train(self, mode: bool = True) -> None:
        """Set distiller's forward mode."""
        super().train(mode)
        if mode and self.teacher_norm_eval:
            for m in self.teacher.modules():
                if isinstance(m, _BatchNorm):
                    m.eval()



# Copyright (c) OpenMMLab. All rights reserved.
from .diff_module_mutator import DiffModuleMutator
from .module_mutator import ModuleMutator
from .one_shot_module_mutator import OneShotModuleMutator

__all__ = ['OneShotModuleMutator', 'DiffModuleMutator', 'ModuleMutator']



_base_ = [
    'mmcls::_base_/datasets/cifar10_bs16.py',
    'mmcls::_base_/schedules/cifar10_bs128.py',
    'mmcls::_base_/default_runtime.py'
]

res34_ckpt_path = 'https://download.openmmlab.com/mmclassification/v0/resnet/resnet34_b16x8_cifar10_20210528-a8aa36a6.pth'  # noqa: E501
model = dict(
    _scope_='mmrazor',
    type='DataFreeDistillation',
    data_preprocessor=dict(
        type='ImgDataPreprocessor',
        # RGB format normalization parameters
        mean=[125.307, 122.961, 113.8575],
        std=[51.5865, 50.847, 51.255],
        # convert image from BGR to RGB
        bgr_to_rgb=False),
    architecture=dict(
        cfg_path='mmcls::resnet/resnet18_8xb16_cifar10.py', pretrained=False),
    teachers=dict(
        res34=dict(
            build_cfg=dict(
                cfg_path='mmcls::resnet/resnet34_8xb16_cifar10.py',
                pretrained=True),
            ckpt_path=res34_ckpt_path)),
    generator=dict(
        type='ZSKTGenerator', img_size=32, latent_dim=256,
        hidden_channels=128),
    distiller=dict(
        type='ConfigurableDistiller',
        student_recorders=dict(
            bb_s1=dict(type='ModuleOutputs', source='backbone.layer1.1.relu'),
            bb_s2=dict(type='ModuleOutputs', source='backbone.layer2.1.relu'),
            bb_s3=dict(type='ModuleOutputs', source='backbone.layer3.1.relu'),
            bb_s4=dict(type='ModuleOutputs', source='backbone.layer4.1.relu'),
            fc=dict(type='ModuleOutputs', source='head.fc')),
        teacher_recorders=dict(
            res34_bb_s1=dict(
                type='ModuleOutputs', source='res34.backbone.layer1.2.relu'),
            res34_bb_s2=dict(
                type='ModuleOutputs', source='res34.backbone.layer2.3.relu'),
            res34_bb_s3=dict(
                type='ModuleOutputs', source='res34.backbone.layer3.5.relu'),
            res34_bb_s4=dict(
                type='ModuleOutputs', source='res34.backbone.layer4.2.relu'),
            res34_fc=dict(type='ModuleOutputs', source='res34.head.fc')),
        distill_losses=dict(
            loss_s1=dict(type='ATLoss', loss_weight=250.0),
            loss_s2=dict(type='ATLoss', loss_weight=250.0),
            loss_s3=dict(type='ATLoss', loss_weight=250.0),
            loss_s4=dict(type='ATLoss', loss_weight=250.0),
            loss_kl=dict(
                type='KLDivergence', loss_weight=2.0, reduction='mean')),
        loss_forward_mappings=dict(
            loss_s1=dict(
                s_feature=dict(
                    from_student=True, recorder='bb_s1', record_idx=1),
                t_feature=dict(
                    from_student=False, recorder='res34_bb_s1', record_idx=1)),
            loss_s2=dict(
                s_feature=dict(
                    from_student=True, recorder='bb_s2', record_idx=1),
                t_feature=dict(
                    from_student=False, recorder='res34_bb_s2', record_idx=1)),
            loss_s3=dict(
                s_feature=dict(
                    from_student=True, recorder='bb_s3', record_idx=1),
                t_feature=dict(
                    from_student=False, recorder='res34_bb_s3', record_idx=1)),
            loss_s4=dict(
                s_feature=dict(
                    from_student=True, recorder='bb_s4', record_idx=1),
                t_feature=dict(
                    from_student=False, recorder='res34_bb_s4', record_idx=1)),
            loss_kl=dict(
                preds_S=dict(from_student=True, recorder='fc'),
                preds_T=dict(from_student=False, recorder='res34_fc')))),
    generator_distiller=dict(
        type='ConfigurableDistiller',
        student_recorders=dict(
            fc=dict(type='ModuleOutputs', source='head.fc')),
        teacher_recorders=dict(
            res34_fc=dict(type='ModuleOutputs', source='res34.head.fc')),
        distill_losses=dict(
            loss_kl=dict(
                type='KLDivergence',
                loss_weight=-2.0,
                reduction='mean',
                teacher_detach=False)),
        loss_forward_mappings=dict(
            loss_kl=dict(
                preds_S=dict(from_student=True, recorder='fc'),
                preds_T=dict(from_student=False, recorder='res34_fc')))),
    student_iter=10)

# model wrapper
model_wrapper_cfg = dict(
    type='mmengine.MMSeparateDistributedDataParallel',
    broadcast_buffers=False,
    find_unused_parameters=True)

# optimizer wrapper
optim_wrapper = dict(
    _delete_=True,
    constructor='mmrazor.SeparateOptimWrapperConstructor',
    architecture=dict(
        optimizer=dict(type='SGD', lr=0.1, weight_decay=0.0005, momentum=0.9)),
    generator=dict(optimizer=dict(type='Adam', lr=1e-3)))
auto_scale_lr = dict(base_batch_size=16)

iter_size = 50

param_scheduler = dict(
    _delete_=True,
    architecture=dict(
        type='MultiStepLR',
        milestones=[100 * iter_size, 200 * iter_size],
        by_epoch=False),
    generator=dict(
        type='MultiStepLR',
        milestones=[100 * iter_size, 200 * iter_size],
        by_epoch=False))

train_cfg = dict(
    _delete_=True, by_epoch=False, max_iters=500 * iter_size, val_interval=250)

train_dataloader = dict(
    batch_size=16, sampler=dict(type='InfiniteSampler', shuffle=True))
val_dataloader = dict(batch_size=16)
val_evaluator = dict(type='Accuracy', topk=(1, 5))

default_hooks = dict(
    logger=dict(type='LoggerHook', interval=50, log_metric_by_epoch=False),
    checkpoint=dict(
        type='CheckpointHook', by_epoch=False, interval=100, max_keep_ckpts=2))

log_processor = dict(by_epoch=False)
# Must set diff_rank_seed to True!
randomness = dict(seed=None, diff_rank_seed=True)



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, Optional

import torch
import torch.nn as nn
import torch.nn.functional as F

from mmrazor.registry import MODELS
from .base_generator import BaseGenerator


@MODELS.register_module()
class DAFLGenerator(BaseGenerator):
    """Generator for DAFL.

    Args:
        img_size (int): The size of generated image.
        latent_dim (int): The dimension of latent data.
        hidden_channels (int): The dimension of hidden channels.
        scale_factor (int, optional): The scale factor for F.interpolate.
            Defaults to 2.
        bn_eps (float, optional): The eps param in bn. Defaults to 0.8.
        leaky_slope (float, optional): The slope param in leaky relu. Defaults
            to 0.2.
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(
        self,
        img_size: int,
        latent_dim: int,
        hidden_channels: int,
        scale_factor: int = 2,
        bn_eps: float = 0.8,
        leaky_slope: float = 0.2,
        init_cfg: Optional[Dict] = None,
    ) -> None:
        super().__init__(
            img_size, latent_dim, hidden_channels, init_cfg=init_cfg)
        self.init_size = self.img_size // (scale_factor**2)
        self.scale_factor = scale_factor
        self.linear = nn.Linear(self.latent_dim,
                                self.hidden_channels * self.init_size**2)

        self.bn1 = nn.BatchNorm2d(self.hidden_channels)
        self.conv_blocks1 = nn.Sequential(
            nn.Conv2d(
                self.hidden_channels,
                self.hidden_channels,
                3,
                stride=1,
                padding=1),
            nn.BatchNorm2d(self.hidden_channels, eps=bn_eps),
            nn.LeakyReLU(leaky_slope, inplace=True),
        )
        self.conv_blocks2 = nn.Sequential(
            nn.Conv2d(
                self.hidden_channels,
                self.hidden_channels // 2,
                3,
                stride=1,
                padding=1),
            nn.BatchNorm2d(self.hidden_channels // 2, eps=bn_eps),
            nn.LeakyReLU(leaky_slope, inplace=True),
            nn.Conv2d(self.hidden_channels // 2, 3, 3, stride=1, padding=1),
            nn.Tanh(), nn.BatchNorm2d(3, affine=False))

    def forward(self,
                data: Optional[torch.Tensor] = None,
                batch_size: int = 1) -> torch.Tensor:
        """Forward function for generator.

        Args:
            data (torch.Tensor, optional): The input data. Defaults to None.
            batch_size (int): Batch size. Defaults to 1.
        """
        batch_data = self.process_latent(data, batch_size)
        img = self.linear(batch_data)
        img = img.view(img.shape[0], self.hidden_channels, self.init_size,
                       self.init_size)
        img = self.bn1(img)
        img = F.interpolate(img, scale_factor=self.scale_factor)
        img = self.conv_blocks1(img)
        img = F.interpolate(img, scale_factor=self.scale_factor)
        img = self.conv_blocks2(img)
        return img



# Copyright (c) OpenMMLab. All rights reserved.
from .hooks import DumpSubnetHook, EstimateResourcesHook
from .optimizers import SeparateOptimWrapperConstructor
from .runner import (AutoSlimValLoop, DartsEpochBasedTrainLoop,
                     DartsIterBasedTrainLoop, EvolutionSearchLoop,
                     GreedySamplerTrainLoop, SelfDistillValLoop,
                     SingleTeacherDistillValLoop, SlimmableValLoop)

__all__ = [
    'SeparateOptimWrapperConstructor', 'DumpSubnetHook',
    'SingleTeacherDistillValLoop', 'DartsEpochBasedTrainLoop',
    'DartsIterBasedTrainLoop', 'SlimmableValLoop', 'EvolutionSearchLoop',
    'GreedySamplerTrainLoop', 'AutoSlimValLoop', 'SelfDistillValLoop',
    'EstimateResourcesHook'
]



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from mmengine.model import BaseModel
from mmengine.structures import BaseDataElement

from mmrazor.registry import MODELS

LossResults = Dict[str, torch.Tensor]
TensorResults = Union[Tuple[torch.Tensor], torch.Tensor]
PredictResults = List[BaseDataElement]
ForwardResults = Union[LossResults, TensorResults, PredictResults]


@MODELS.register_module()
class BaseAlgorithm(BaseModel):
    """Base class for algorithms.

    BaseAlgorithm inherit from BaseModel. BaseModel implements the basic
    functions of the algorithmic model, such as weights initialize,
    batch inputs preprocess(see more information in
    :class:`BaseDataPreprocessor`), parse losses, and update model parameters.
    More details of BaseModel could see docs for :class:`BaseModel`.

    :obj:`BaseAlgorithm` forward just is a wrapper of :obj:`BaseModel` forward.
    Various compression algorithms can be implemented by inheriting
    BaseAlgorithm.

    Subclasses inherit from BaseAlgorithm only need to override the
    :meth:`loss`, which implements the logic to calculate loss, then
    can be trained in the runner.

    Args:
        architecture (dict | :obj:`BaseModel`): The config of
            :class:`BaseModel` or built model.
        data_preprocessor (dict | torch.nn.Module | None): The pre-process
            config of :class:`BaseDataPreprocessor`. Defaults to None.
        init_cfg (dict): The weight initialized config for
            :class:`BaseModule`.

    Note:
        If `data_preprocessor` is None, :obj:`BaseAlgorithm` will set
        `data_preprocessor` to model's `data_preprocessor`.


    Attributes:
        architecture (:obj:`BaseModel`): Model that needs to be compressed.
        data_preprocessor (:obj:`BaseDataPreprocessor`): Used for
            pre-processing data sampled by dataloader to the format accepted by
            :meth:`forward`.
        init_cfg (dict, optional): Initialization config dict.
    """

    def __init__(self,
                 architecture: Union[BaseModel, Dict],
                 data_preprocessor: Optional[Union[Dict, nn.Module]] = None,
                 init_cfg: Optional[Dict] = None) -> None:

        # super().__init__() needs built data_preprocessor, so
        # build model first.
        if isinstance(architecture, Dict):
            architecture = MODELS.build(architecture)

        if not isinstance(architecture, BaseModel):
            raise TypeError('architecture should be a `dict` or '
                            f'`BaseModel` instance, but got '
                            f'{type(architecture)}')

        # If `data_preprocessor` is None, there will set
        # `data_preprocessor` to model's `data_preprocessor`.
        if data_preprocessor is None:
            # use model's data_preprocessor
            data_preprocessor = getattr(architecture, 'data_preprocessor',
                                        None)
        super().__init__(data_preprocessor, init_cfg)

        # Cannot assign module before Module.__init__()
        self.architecture = architecture

    def forward(self,
                inputs: torch.Tensor,
                data_samples: Optional[List[BaseDataElement]] = None,
                mode: str = 'tensor') -> ForwardResults:
        """Returns losses or predictions of training, validation, testing, and
        simple inference process.

        ``forward`` method of BaseModel is an abstract method, its subclasses
        must implement this method.

        Accepts ``batch_inputs`` and ``data_samples`` processed by
        :attr:`data_preprocessor`, and returns results according to mode
        arguments.

        During non-distributed training, validation, and testing process,
        ``forward`` will be called by ``BaseModel.train_step``,
        ``BaseModel.val_step`` and ``BaseModel.val_step`` directly.

        During distributed data parallel training process,
        ``MMSeparateDistributedDataParallel.train_step`` will first call
        ``DistributedDataParallel.forward`` to enable automatic
        gradient synchronization, and then call ``forward`` to get training
        loss.

        Args:
            batch_inputs (torch.Tensor): batch input tensor collated by
                :attr:`data_preprocessor`.
            data_samples (List[BaseDataElement], optional):
                data samples collated by :attr:`data_preprocessor`.
            mode (str): mode should be one of ``loss``, ``predict`` and
                ``tensor``
                - ``loss``: Called by ``train_step`` and return loss ``dict``
                  used for logging
                - ``predict``: Called by ``val_step`` and ``test_step``
                  and return list of ``BaseDataElement`` results used for
                  computing metric.
                - ``tensor``: Called by custom use to get ``Tensor`` type
                  results.

        Returns:
            ForwardResults:
                - If ``mode == loss``, return a ``dict`` of loss tensor used
                  for backward and logging.
                - If ``mode == predict``, return a ``list`` of
                  :obj:`BaseDataElement` for computing metric
                  and getting inference result.
                - If ``mode == tensor``, return a tensor or ``tuple`` of tensor
                  or ``dict of tensor for custom use.
        """
        if mode == 'loss':
            return self.loss(inputs, data_samples)
        elif mode == 'tensor':
            return self._forward(inputs, data_samples)
        elif mode == 'predict':
            return self._predict(inputs, data_samples)
        else:
            raise RuntimeError(f'Invalid mode "{mode}". '
                               'Only supports loss, predict and tensor mode')

    def loss(
        self,
        inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
    ) -> LossResults:
        """Calculate losses from a batch of inputs and data samples."""
        return self.architecture(inputs, data_samples, mode='loss')

    def _forward(
        self,
        inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
    ) -> TensorResults:
        """Network forward process."""
        return self.architecture(inputs, data_samples, mode='tensor')

    def _predict(
        self,
        inputs: torch.Tensor,
        data_samples: Optional[List[BaseDataElement]] = None,
    ) -> PredictResults:
        """Predict results from a batch of inputs and data samples with post-
        processing."""
        return self.architecture(inputs, data_samples, mode='predict')



# Copyright (c) OpenMMLab. All rights reserved.
import warnings
from typing import Any, Dict, List, Optional

from torch.nn import Module

from mmrazor.registry import MODELS
from ...mutables import OneShotMutableChannel
from .channel_mutator import ChannelMutator


@MODELS.register_module()
class OneShotChannelMutator(ChannelMutator):
    """One-shot channel mutable based channel mutator.

    Args:
        mutable_cfg (dict): The config for the channel mutable.
        tracer_cfg (dict): The config for the model tracer. We Trace the
            topology of a given model with the tracer.
        skip_prefixes (List[str] | Optional): The module whose name start with
            a string in skip_prefixes will not be pruned.
        init_cfg (dict, optional): The config to control the initialization.
    """

    def __init__(self,
                 mutable_cfg: Dict,
                 tracer_cfg: Optional[Dict] = None,
                 skip_prefixes: Optional[List[str]] = None,
                 init_cfg: Optional[Dict] = None) -> None:
        super().__init__(mutable_cfg, tracer_cfg, skip_prefixes, init_cfg)

    def sample_choices(self):
        """Sample a choice that records a selection from the search space.

        Returns:
            dict: Record the information to build the subnet from the supernet.
                Its keys are the properties ``group_idx`` in the channel
                mutator's ``search_groups``, and its values are the sampled
                choice.
        """
        choice_dict = dict()
        for group_idx, mutables in self.search_groups.items():
            choice_dict[group_idx] = mutables[0].sample_choice()
        return choice_dict

    def set_choices(self, choice_dict: Dict[int, Any]) -> None:
        """Set current subnet according to ``choice_dict``.

        Args:
            choice_dict (Dict[int, Any]): Choice dict.
        """
        for group_idx, choice in choice_dict.items():
            mutables = self.search_groups[group_idx]
            for mutable in mutables:
                mutable.current_choice = choice

    def set_max_choices(self) -> None:
        """Set the channel numbers of each layer to maximum."""
        for mutables in self.search_groups.values():
            for mutable in mutables:
                mutable.current_choice = mutable.max_choice

    def set_min_choices(self) -> None:
        """Set the channel numbers of each layer to minimum."""
        for mutables in self.search_groups.values():
            for mutable in mutables:
                mutable.current_choice = mutable.min_choice

    # todo: check search gorups
    def build_search_groups(self, supernet: Module):
        """Build `search_groups`. The mutables in the same group should be
        pruned together.

        Examples:
            >>> class ResBlock(nn.Module):
            ...     def __init__(self) -> None:
            ...         super().__init__()
            ...
            ...         self.op1 = nn.Conv2d(3, 8, 1)
            ...         self.bn1 = nn.BatchNorm2d(8)
            ...         self.op2 = nn.Conv2d(8, 8, 1)
            ...         self.bn2 = nn.BatchNorm2d(8)
            ...         self.op3 = nn.Conv2d(8, 8, 1)
            ...
            ...      def forward(self, x):
            ...         x1 = self.bn1(self.op1(x))
            ...         x2 = self.bn2(self.op2(x1))
            ...         x3 = self.op3(x2 + x1)
            ...         return x3

            >>> class ToyPseudoLoss:
            ...
            ...     def __call__(self, model):
            ...         pseudo_img = torch.rand(2, 3, 16, 16)
            ...         pseudo_output = model(pseudo_img)
            ...         return pseudo_output.sum()

            >>> mutator = OneShotChannelMutator(
            ...     tracer_cfg=dict(type='BackwardTracer',
            ...         loss_calculator=ToyPseudoLoss()),
            ...     mutable_cfg=dict(type='OneShotMutableChannel',
            ...         candidate_choices=[4 / 8, 1.0], candidate_mode='ratio')

            >>> model = ResBlock()
            >>> mutator.prepare_from_supernet(model)
            >>> mutator.search_groups
            {0: [OneShotMutableChannel(name=op2, ...), # mutable out
                 OneShotMutableChannel(name=op1, ...), # mutable out
                 OneShotMutableChannel(name=op3, ...), # mutable in
                 OneShotMutableChannel(name=op2, ...), # mutable in
                 OneShotMutableChannel(name=bn2, ...), # mutable out
                 OneShotMutableChannel(name=bn1, ...)] # mutable out
            }
        """
        groups = self.find_same_mutables(supernet)

        search_groups = dict()
        group_idx = 0
        for group in groups.values():
            is_skip = False
            for mutable in group:
                if self.is_skip_pruning(mutable.name, self.skip_prefixes):
                    warnings.warn(f'Group {group} is not searchable due to'
                                  f' skip_prefixes: {self.skip_prefixes}')
                    is_skip = True
                    break
            if not is_skip:
                search_groups[group_idx] = group
                group_idx += 1

        return search_groups

    def mutable_class_type(self):
        """One-shot channel mutable class type.

        Returns:
            Type[OneShotMutableModule]: Class type of one-shot mutable.
        """
        return OneShotMutableChannel



# Copyright (c) OpenMMLab. All rights reserved.
from collections import UserList
from typing import Any, Dict, List, Optional, Tuple, Union


class Candidates(UserList):
    """The data structure of sampled candidate. The format is [(any, float),
    (any, float), ...].

    Examples:
        >>> candidates = Candidates()
        >>> subnet_1 = {'choice_1': 'layer_1', 'choice_2': 'layer_2'}
        >>> candidates.append(subnet_1)
        >>> candidates
        [({'choice_1': 'layer_1', 'choice_2': 'layer_2'}, 0.0)]
        >>> candidates.set_score(0, 0.9)
        >>> candidates
        [({'choice_1': 'layer_1', 'choice_2': 'layer_2'}, 0.9)]
        >>> subnet_2 = {'choice_3': 'layer_3', 'choice_4': 'layer_4'}
        >>> candidates.append((subnet_2, 0.5))
        >>> candidates
        [({'choice_1': 'layer_1', 'choice_2': 'layer_2'}, 0.9),
        ({'choice_3': 'layer_3', 'choice_4': 'layer_4'}, 0.5)]
        >>> candidates.subnets
        [{'choice_1': 'layer_1', 'choice_2': 'layer_2'},
        {'choice_3': 'layer_3', 'choice_4': 'layer_4'}]
        >>> candidates.scores
        [0.9, 0.5]
    """
    _format_return = Union[Tuple[Any, float], List[Tuple[Any, float]]]

    def __init__(self, initdata: Optional[Any] = None):
        self.data = []
        if initdata is not None:
            initdata = self._format(initdata)
            if isinstance(initdata, list):
                self.data = initdata
            else:
                self.data.append(initdata)

    @property
    def scores(self) -> List[float]:
        """The scores of candidates."""
        return [item[1] for item in self.data]

    @property
    def subnets(self) -> List[Dict]:
        """The subnets of candidates."""
        return [item[0] for item in self.data]

    def _format(self, data: Any) -> _format_return:
        """Transform [any, ...] to [tuple(any, float), ...] Transform any to
        tuple(any, float)."""

        def _format_item(item: Any):
            """Transform any to tuple(any, float)."""
            if isinstance(item, tuple):
                return (item[0], float(item[1]))
            else:
                return (item, 0.)

        if isinstance(data, UserList):
            return [_format_item(i) for i in data.data]

        elif isinstance(data, list):
            return [_format_item(i) for i in data]

        else:
            return _format_item(data)

    def append(self, item: Any) -> None:
        """Append operation."""
        item = self._format(item)
        self.data.append(item)

    def insert(self, i: int, item: Any) -> None:
        """Insert operation."""
        item = self._format(item)
        self.data.insert(i, item)

    def extend(self, other: Any) -> None:
        """Extend operation."""
        other = self._format(other)
        if isinstance(other, list):
            self.data.extend(other)
        else:
            self.data.extend([other])

    def set_score(self, i: int, score: float) -> None:
        """Set score to the specified subnet by index."""
        self.data[i] = (self.data[i][0], float(score))



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Dict

from torch.nn.modules.batchnorm import _BatchNorm

from mmrazor.models.architectures import SwitchableBatchNorm2d


def switchable_bn_converter(module: _BatchNorm, in_channels_cfg: Dict,
                            out_channels_cfg: Dict) -> SwitchableBatchNorm2d:
    """Convert a _BatchNorm module to a SwitchableBatchNorm2d.

    Args:
        module (:obj:`torch.nn.GroupNorm`): The original BatchNorm module.
        num_channels_cfg (Dict): Config related to `num_features`.
    """
    switchable_bn = SwitchableBatchNorm2d(
        num_features_cfg=in_channels_cfg,
        eps=module.eps,
        momentum=module.momentum,
        affine=module.affine,
        track_running_stats=module.track_running_stats)
    return switchable_bn



# Copyright (c) OpenMMLab. All rights reserved.
from typing import Any, Dict, List, Tuple

import torch.nn
from mmengine.dist import broadcast_object_list, is_main_process

from mmrazor.registry import TASK_UTILS
from .base_estimator import BaseEstimator
from .counters import (get_model_complexity_info, params_units_convert,
                       repeat_measure_inference_speed)


@TASK_UTILS.register_module()
class ResourceEstimator(BaseEstimator):
    """Estimator for calculating the resources consume.

    Args:
        default_shape (tuple): Input data's default shape, for calculating
            resources consume. Defaults to (1, 3, 224, 224)
        units (str): Resource units. Defaults to 'M'.
        disabled_counters (list): List of disabled spec op counters.
            Defaults to None.
        NOTE: disabled_counters contains the op counter class names
              in estimator.op_counters that require to be disabled,
              such as 'ConvCounter', 'BatchNorm2dCounter', ...

    Examples:
        >>> # direct calculate resource consume of nn.Conv2d
        >>> conv2d = nn.Conv2d(3, 32, 3)
        >>> estimator = ResourceEstimator()
        >>> estimator.estimate(
        ...     model=conv2d,
        ...     resource_args=dict(input_shape=(1, 3, 64, 64)))
        {'flops': 3.444, 'params': 0.001, 'latency': 0.0}

        >>> # calculate resources of custom modules
        >>> class CustomModule(nn.Module):
        ...
        ...    def __init__(self) -> None:
        ...        super().__init__()
        ...
        ...    def forward(self, x):
        ...        return x
        ...
        >>> @TASK_UTILS.register_module()
        ... class CustomModuleCounter(BaseCounter):
        ...
        ...    @staticmethod
        ...    def add_count_hook(module, input, output):
        ...        module.__flops__ += 1000000
        ...        module.__params__ += 700000
        ...
        >>> model = CustomModule()
        >>> estimator.estimate(
        ...     model=model,
        ...     resource_args=dict(input_shape=(1, 3, 64, 64)))
        {'flops': 1.0, 'params': 0.7, 'latency': 0.0}
        ...
        >>> # calculate resources of custom modules with disable_counters
        >>> estimator.estimate(
        ...     model=model,
        ...     resource_args=dict(
        ...         input_shape=(1, 3, 64, 64),
        ...         disabled_counters=['CustomModuleCounter']))
        {'flops': 0.0, 'params': 0.0, 'latency': 0.0}

        >>> # calculate resources of mmrazor.models
        NOTE: check 'EstimateResourcesHook' in
              mmrazor.engine.hooks.estimate_resources_hook for details.
    """

    def __init__(self,
                 default_shape: Tuple = (1, 3, 224, 224),
                 units: str = 'M',
                 disabled_counters: List[str] = [],
                 as_strings: bool = False,
                 measure_inference: bool = False):
        super().__init__(default_shape, units, disabled_counters, as_strings,
                         measure_inference)

    def estimate(
        self, model: torch.nn.Module, resource_args: Dict[str, Any] = dict()
    ) -> Dict[str, Any]:
        """Estimate the resources(flops/params/latency) of the given model.

        Args:
            model: The measured model.
            resource_args (Dict[str, float]): Args for resources estimation.
            NOTE: resource_args have the same items() as the init cfgs.

        Returns:
            Dict[str, str]): A dict that containing resource results(flops,
                params and latency).
        """
        resource_metrics = dict()
        if is_main_process():
            measure_inference = resource_args.pop('measure_inference', False)
            if 'input_shape' not in resource_args.keys():
                resource_args['input_shape'] = self.default_shape
            if 'disabled_counters' not in resource_args.keys():
                resource_args['disabled_counters'] = self.disabled_counters
            model.eval()
            flops, params = get_model_complexity_info(model, **resource_args)
            if measure_inference:
                latency = repeat_measure_inference_speed(
                    model, resource_args, max_iter=100, repeat_num=2)
            else:
                latency = 0.0
            as_strings = resource_args.get('as_strings', self.as_strings)
            if as_strings and self.units is not None:
                raise ValueError('Set units to None, when as_trings=True.')
            if self.units is not None:
                flops = params_units_convert(flops, self.units)
                params = params_units_convert(params, self.units)
            resource_metrics.update({
                'flops': flops,
                'params': params,
                'latency': latency
            })
            results = [resource_metrics]
        else:
            results = [None]  # type: ignore

        broadcast_object_list(results)

        return results[0]

    def estimate_spec_modules(
        self, model: torch.nn.Module, resource_args: Dict[str, Any] = dict()
    ) -> Dict[str, float]:
        """Estimate the resources(flops/params/latency) of the spec modules.

        Args:
            model: The measured model.
            resource_args (Dict[str, float]): Args for resources estimation.
            NOTE: resource_args have the same items() as the init cfgs.

        Returns:
            Dict[str, float]): A dict that containing resource results(flops,
                params) of each modules in resource_args['spec_modules'].
        """
        assert 'spec_modules' in resource_args, \
            'spec_modules is required when calling estimate_spec_modules().'

        resource_args.pop('measure_inference', False)
        if 'input_shape' not in resource_args.keys():
            resource_args['input_shape'] = self.default_shape
        if 'disabled_counters' not in resource_args.keys():
            resource_args['disabled_counters'] = self.disabled_counters

        model.eval()
        spec_modules_resources = get_model_complexity_info(
            model, **resource_args)

        return spec_modules_resources



# Copyright (c) OpenMMLab. All rights reserved.
from .channel_mutator import ChannelMutator
from .one_shot_channel_mutator import OneShotChannelMutator
from .slimmable_channel_mutator import SlimmableChannelMutator

__all__ = [
    'ChannelMutator', 'OneShotChannelMutator', 'SlimmableChannelMutator'
]



# Copyright (c) OpenMMLab. All rights reserved.
import copy
from typing import Dict, List, Optional, Tuple, Union

import torch
import torch.nn as nn
from mmcv.cnn import build_activation_layer, build_norm_layer
from torch import Tensor

from mmrazor.registry import MODELS


class FactorizedReduce(nn.Module):
    """Reduce feature map size by factorized pointwise (stride=2).

    Args:
        in_channels (int): number of channels of input tensor.
        out_channels (int): number of channels of output tensor.
        act_cfg (Dict): config to build activation layer.
        norm_cfg (Dict): config to build normalization layer.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        act_cfg: Dict = dict(type='ReLU'),
        norm_cfg: Dict = dict(type='BN')
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.act_cfg = act_cfg
        self.norm_cfg = norm_cfg
        self.relu = build_activation_layer(self.act_cfg)
        self.conv1 = nn.Conv2d(
            self.in_channels,
            self.out_channels // 2,
            1,
            stride=2,
            padding=0,
            bias=False)
        self.conv2 = nn.Conv2d(
            self.in_channels,
            self.out_channels // 2,
            1,
            stride=2,
            padding=0,
            bias=False)
        self.bn = build_norm_layer(self.norm_cfg, self.out_channels)[1]

    def forward(self, x: Tensor) -> Tensor:
        """Forward with factorized reduce."""
        x = self.relu(x)
        out = torch.cat([self.conv1(x), self.conv2(x[:, :, 1:, 1:])], dim=1)
        out = self.bn(out)
        return out


class StandardConv(nn.Module):
    """Standard Convolution in Darts. Basic structure is ReLU-Conv-BN.

    Args:
        in_channels (int): number of channels of input tensor.
        out_channels (int): number of channels of output tensor.
        kernel_size (Union[int, Tuple]): size of the convolving kernel.
        stride (Union[int, Tuple]): controls the stride for the
            cross-correlation, a single number or a one-element tuple.
            Default to 1.
        padding (Union[str, int, Tuple]): Padding added to both sides
            of the input. Default to 0.
        act_cfg (Dict): config to build activation layer.
        norm_cfg (Dict): config to build normalization layer.
    """

    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        kernel_size: Union[int, Tuple],
        stride: Union[int, Tuple] = 1,
        padding: Union[str, int, Tuple] = 0,
        act_cfg: Dict = dict(type='ReLU'),
        norm_cfg: Dict = dict(type='BN')
    ) -> None:
        super().__init__()
        self.in_channels = in_channels
        self.out_channels = out_channels
        self.kernel_size = kernel_size
        self.stride = stride
        self.padding = padding
        self.act_cfg = act_cfg
        self.norm_cfg = norm_cfg
        self.net = nn.Sequential(
            build_activation_layer(self.act_cfg),
            nn.Conv2d(
                self.in_channels,
                self.out_channels,
                self.kernel_size,
                self.stride,
                self.padding,
                bias=False),
            build_norm_layer(self.norm_cfg, self.out_channels)[1])

    def forward(self, x: Tensor) -> Tensor:
        """Forward the standard convolution."""
        return self.net(x)


class Node(nn.Module):
    """Node structure of DARTS.

    Args:
        node_id (str): key of the node.
        num_prev_nodes (int): number of previous nodes.
        channels (int): number of channels of current node.
        num_downsample_nodes (int): index of downsample node.
        mutable_cfg (Dict): config of `DiffMutableModule`.
        route_cfg (Dict): config of `DiffChoiceRoute`.
    """

    def __init__(self, node_id: str, num_prev_nodes: int, channels: int,
                 num_downsample_nodes: int, mutable_cfg: Dict,
                 route_cfg: Dict) -> None:
        super().__init__()
        edges = nn.ModuleDict()
        for i in range(num_prev_nodes):
            stride = 2 if i < num_downsample_nodes else 1
            edge_id = f'{node_id}_p{i}'

            module_kwargs = dict(
                in_channels=channels,
                out_channels=channels,
                stride=stride,
            )

            mutable_cfg.update(module_kwargs=module_kwargs)
            mutable_cfg.update(alias=edge_id)
            edges.add_module(edge_id, MODELS.build(mutable_cfg))

        route_cfg.update(alias=node_id)
        route_cfg.update(edges=edges)
        self.route = MODELS.build(route_cfg)

    def forward(self, prev_nodes: Union[List[Tensor],
                                        Tuple[Tensor]]) -> Tensor:
        """Forward with the previous nodes list."""
        return self.route(prev_nodes)


class Cell(nn.Module):
    """Darts cell structure.

    Args:
        num_nodes (int): number of nodes.
        channels (int): number of channels of current cell.
        prev_channels (int): number of channel of previous input.
        prev_prev_channels (int): number of channel of previous previous input.
        reduction (bool): whether to reduce the feature map size.
        prev_reduction (bool): whether to reduce the previous feature map size.
        mutable_cfg (Optional[Dict]): config of `DiffMutableModule`.
        route_cfg (Optional[Dict]): config of `DiffChoiceRoute`.
        act_cfg (Dict): config to build activation layer.
            Defaults to dict(type='ReLU').
        norm_cfg (Dict): config to build normalization layer.
            Defaults to dict(type='BN').
    """

    def __init__(
            self,
            num_nodes: int,
            channels: int,
            prev_channels: int,
            prev_prev_channels: int,
            reduction: bool,
            prev_reduction: bool,
            mutable_cfg: Dict,
            route_cfg: Dict,
            act_cfg: Dict = dict(type='ReLU'),
            norm_cfg: Dict = dict(type='BN'),
    ) -> None:

        super().__init__()
        self.act_cfg = act_cfg
        self.norm_cfg = norm_cfg
        self.reduction = reduction
        self.num_nodes = num_nodes

        # If previous cell is reduction cell, current input size does not match
        # with output size of cell[k-2]. So the output[k-2] should be reduced
        # by preprocessing.
        if prev_reduction:
            self.preproc0 = FactorizedReduce(prev_prev_channels, channels,
                                             self.act_cfg, self.norm_cfg)
        else:
            self.preproc0 = StandardConv(prev_prev_channels, channels, 1, 1, 0,
                                         self.act_cfg, self.norm_cfg)
        self.preproc1 = StandardConv(prev_channels, channels, 1, 1, 0,
                                     self.act_cfg, self.norm_cfg)

        # generate dag
        self.nodes = nn.ModuleList()
        for depth in range(2, self.num_nodes + 2):
            if reduction:
                node_id = f'reduce_n{depth}'
                num_downsample_nodes = 2
            else:
                node_id = f'normal_n{depth}'
                num_downsample_nodes = 0
            self.nodes.append(
                Node(node_id, depth, channels, num_downsample_nodes,
                     mutable_cfg, route_cfg))

    def forward(self, s0: Tensor, s1: Tensor) -> Tensor:
        """Forward with the outputs of previous previous cell and previous
        cell."""
        tensors = [self.preproc0(s0), self.preproc1(s1)]
        for node in self.nodes:
            cur_tensor = node(tensors)
            tensors.append(cur_tensor)

        return torch.cat(tensors[2:], dim=1)


class AuxiliaryModule(nn.Module):
    """Auxiliary head in 2/3 place of network to let the gradient flow well.

    Args:
        in_channels (int): number of channels of inputs.
        base_channels (int): number of middle channels of the auxiliary module.
        out_channels (int): number of channels of outputs.
        norm_cfg (Dict): config to build normalization layer.
            Defaults to dict(type='BN').
    """

    def __init__(self,
                 in_channels: int,
                 base_channels: int,
                 out_channels: int,
                 norm_cfg: Dict = dict(type='BN')) -> None:
        super().__init__()
        self.norm_cfg = norm_cfg
        self.net = nn.Sequential(
            nn.ReLU(),
            nn.AvgPool2d(5, stride=2, padding=0,
                         count_include_pad=False),  # 2x2 out
            nn.Conv2d(in_channels, base_channels, kernel_size=1, bias=False),
            build_norm_layer(self.norm_cfg, base_channels)[1],
            nn.ReLU(inplace=True),
            nn.Conv2d(base_channels, out_channels, kernel_size=2,
                      bias=False),  # 1x1 out
            build_norm_layer(self.norm_cfg, out_channels)[1],
            nn.ReLU(inplace=True))

    def forward(self, x: Tensor) -> Tensor:
        """Forward the auxiliary module."""
        return self.net(x)


@MODELS.register_module()
class DartsBackbone(nn.Module):
    """Backbone of Differentiable Architecture Search (DARTS).

    Args:
        in_channels (int): number of channels of input tensor.
        base_channels (int): number of middle channels.
        mutable_cfg (Optional[Dict]): config of `DiffMutableModule`.
        route_cfg (Optional[Dict]): config of `DiffChoiceRoute`.
        num_layers (Optional[int]): number of layers.
            Defaults to 8.
        num_nodes (Optional[int]): number of nodes.
            Defaults to 4.
        stem_multiplier (Optional[int]): multiplier for stem.
            Defaults to 3.
        out_indices (tuple, optional): output indices for auxliary module.
            Defaults to (7, ).
        auxliary (bool, optional): whether use auxliary module.
            Defaults to False.
        aux_channels (Optional[int]): number of middle channels of
            auxliary module. Defaults to None.
        aux_out_channels (Optional[int]): number of output channels of
            auxliary module. Defaults to None.
        act_cfg (Dict): config to build activation layer.
            Defaults to dict(type='ReLU').
        norm_cfg (Dict): config to build normalization layer.
            Defaults to dict(type='BN').
    """

    def __init__(
            self,
            in_channels: int,
            base_channels: int,
            mutable_cfg: Dict,
            route_cfg: Dict,
            num_layers: int = 8,
            num_nodes: int = 4,
            stem_multiplier: int = 3,
            out_indices: Union[Tuple, List] = (7, ),
            auxliary: bool = False,
            aux_channels: Optional[int] = None,
            aux_out_channels: Optional[int] = None,
            act_cfg: Dict = dict(type='ReLU'),
            norm_cfg: Dict = dict(type='BN'),
    ) -> None:
        super().__init__()

        self.in_channels = in_channels
        self.base_channels = base_channels
        self.num_layers = num_layers
        self.num_nodes = num_nodes
        self.stem_multiplier = stem_multiplier
        self.out_indices = out_indices
        assert self.out_indices[-1] == self.num_layers - 1
        if auxliary:
            assert aux_channels is not None
            assert aux_out_channels is not None
            self.aux_channels = aux_channels
            self.aux_out_channels = aux_out_channels
            self.auxliary_indice = 2 * self.num_layers // 3

        else:
            self.auxliary_indice = -1
        self.norm_cfg = norm_cfg
        self.act_cfg = act_cfg
        self.out_channels = self.stem_multiplier * self.base_channels
        stem_norm_cfg = copy.deepcopy(self.norm_cfg)
        stem_norm_cfg.update(dict(affine=True))
        self.stem = nn.Sequential(
            nn.Conv2d(
                self.in_channels, self.out_channels, 3, 1, 1, bias=False),
            build_norm_layer(self.norm_cfg, self.out_channels)[1])

        # for the first cell, stem is used for both s0 and s1
        # prev_prev_channels and prev_channels is output channel size,
        # but c_cur is input channel size.
        prev_prev_channels = self.out_channels
        prev_channels = self.out_channels
        self.out_channels = self.base_channels

        self.cells = nn.ModuleList()
        prev_reduction, reduction = False, False
        for i in range(self.num_layers):
            prev_reduction, reduction = reduction, False
            # Reduce featuremap size and double channels in 1/3
            # and 2/3 layer.
            if i in [self.num_layers // 3, 2 * self.num_layers // 3]:
                self.out_channels *= 2
                reduction = True

            cell = Cell(self.num_nodes, self.out_channels, prev_channels,
                        prev_prev_channels, reduction, prev_reduction,
                        mutable_cfg, route_cfg, self.act_cfg, self.norm_cfg)
            self.cells.append(cell)

            prev_prev_channels = prev_channels
            prev_channels = self.out_channels * self.num_nodes

            if i == self.auxliary_indice:
                self.auxliary_module = AuxiliaryModule(prev_channels,
                                                       self.aux_channels,
                                                       self.aux_out_channels,
                                                       self.norm_cfg)

    def forward(self, x: Tensor) -> Tensor:
        """Forward the darts backbone."""
        outs = []
        s0 = s1 = self.stem(x)
        for i, cell in enumerate(self.cells):
            s0, s1 = s1, cell(s0, s1)
            if i in self.out_indices:
                outs.append(s1)
            if i == self.auxliary_indice and self.training:
                aux_feature = self.auxliary_module(s1)
                outs.insert(0, aux_feature)

        return tuple(outs)
